r"""
Manage Windows Local Group Policy
=================================

.. versionadded:: 2016.11.0

This state module allows you to configure local Group Policy on Windows. You
can ensure the setting of a single policy or multiple policies in one pass.

Single policies must specify the policy name, the setting, and the policy class
(Machine/User/Both). Here are some examples for setting a single policy setting.

Example single policy configuration:

.. code-block:: yaml

    Ensure Account Lockout Duration:
      lgpo.set:
        - name: Account lockout duration
        - setting: 90
        - policy_class: Machine

Example using abbreviated form:

.. code-block:: yaml

    Account lockout duration:
      lgpo.set:
        - setting: 120
        - policy_class: Machine

It is also possible to set multiple policies in a single state. This is done by
setting the settings under either `computer_policy` or `user_policy`. Here are
some examples for setting multiple policy settings in a single state.

Multiple policy configuration

.. code-block:: yaml

    Company Local Group Policy:
        lgpo.set:
            - computer_policy:
                Deny log on locally:
                  - Guest
                Account lockout duration: 120
                Account lockout threshold: 10
                Reset account lockout counter after: 120
                Enforce password history: 24
                Maximum password age: 60
                Minimum password age: 1
                Minimum password length: 14
                Password must meet complexity requirements: Enabled
                Store passwords using reversible encryption: Disabled
                Configure Automatic Updates:
                    Configure automatic updating: 4 - Auto download and schedule the intsall
                    Scheduled install day: 7 - Every Saturday
                    Scheduled install time: 17:00
                Specify intranet Microsoft update service location:
                    Set the intranet update service for detecting updates: http://mywsus
                    Set the intranet statistics server: http://mywsus
            - user_policy:
                Do not process the legacy run list: Enabled

.. code-block:: text

    server_policy:
      lgpo.set:
        - computer_policy:
            Maximum password age: 60
            Minimum password age: 1
            Minimum password length: 14
            Account lockout duration: 120
            Account lockout threshold: 10
            Reset account lockout counter after: 120
            Manage auditing and security log:
              - "BUILTIN\\Administrators"
            Replace a process level token:
              - "NT AUTHORITY\\NETWORK SERVICE"
              - "NT AUTHORITY\\LOCAL SERVICE"
            "Accounts: Guest account status": Disabled
            "Accounts: Rename guest account": Not_4_U
            "Audit: Audit the use of Backup and Restore privilege": Enabled
            "Interactive logon: Do not display last user name": Enabled
            "Network\\DNS Client\\Dynamic update": Disabled
            "System\\Logon\\Do not display the Getting Started welcome screen at logon": Enabled
            "Windows Components\\Remote Desktop Services\\Remote Desktop Session Host\\Connections\\Select RDP transport protocols":
                "Select Transport Type": "Use both UDP and TCP"
            "Windows Components\\Windows Update\\Allow Automatic Updates immediate installation": Enabled
            "Windows Components\\Windows Update\\Allow non-administrators to receive update notifications": Disabled
            "Windows Components\\Windows Update\\Always automatically restart at the scheduled time":
                "The restart timer will give users this much time to save their work (minutes)": 15
            "Windows Components\\Windows Update\\Automatic Updates detection frequency":
                "Check for updates at the following interval (hours)": 1
            "Windows Components\\Windows Update\\Configure Automatic Updates":
                "Configure automatic updating": 4 - Auto download and schedule the install
                "Install during automatic maintenance": False
                "Scheduled install day": 7 - Every Saturday
                "Scheduled install time": "17:00"
            "Windows Components\\Windows Update\\Delay Restart for scheduled installations":
                "Wait the following period before proceeding with a scheduled restart (minutes)": 1
            "Windows Components\\Windows Update\\No auto-restart with logged on users for scheduled automatic updates installations": Disabled
            "Windows Components\\Windows Update\\Re-prompt for restart with scheduled installations":
                "Wait the following period before prompting again with a scheduled restart (minutes)": 30
            "Windows Components\\Windows Update\\Reschedule Automatic Updates scheduled installations": Disabled
            "Windows Components\\Windows Update\\Specify intranet Microsoft update service location":
                "Set the intranet update service for detecting updates": http://mywsus
                "Set the intranet statistics server": http://mywsus
        - cumulative_rights_assignments: True

    Some policy settings can't be set on their own an require that other policy
    settings are set at the same time. It can be difficult to figure out what
    additional settings need to be applied. The easiest way to do this is to
    modify the setting manually using the Group Policy Editor (`gpedit.msc`) on
    the machine. Then `get` the policy settings configured on that machine. Use
    the following command:

    .. code-block:: bash

        salt-call --local lgpo.get machine

    For example, if I want to set the Windows Update settings for a Windows
    Server 2016 machine I would go into the Group Policy Editor (`gpedit.msc`)
    and configure the group policy. That policy can be found at: Computer
    Configuration -> Administrative Templates -> Windows Components -> Windows
    Update -> Configure Automatic Updates. You have the option to "Enable" the
    policy and set some configuration options. In this example, just click
    "Enable" and accept the default configuration options. Click "OK" to apply
    the setting.

    Now run the `get` command as shown above. You will find the following in
    the minion return:

    .. code-block:: bash

        Windows Components\Windows Update\Configure Automatic Updates:
            ----------
            Configure automatic updating:
                3 - Auto download and notify for install
            Install during automatic maintenance:
                False
            Install updates for other Microsoft products:
                False
            Scheduled install day:
                0 - Every day
            Scheduled install time:
                03:00

    This shows you that to enable the "Configure Automatic Updates" policy you
    also have to configure the following settings:

    - Configure automatic updating
    - Install during automatic maintenance
    - Install updates for other Microsoft products
    - Scheduled install day
    - Scheduled install time

    So, if you were writing a state for the above policy, it would look like
    this:

    .. code-block:: bash

        configure_windows_update_settings:
          lgpo.set:
            - computer_policy:
                Configure Automatic Updates:
                  Configure automatic updating: 3 - Auto download and notify for install
                  Install during automatic maintenance: False
                  Install updates for other Microsoft products: False
                  Scheduled install day: 0 - Every day
                  Scheduled install time: 03:00

    .. note::

        It is important that you put names of policies and settings exactly as
        they are displayed in the return. That includes capitalization and
        punctuation such as periods, dashes, etc. This rule applies to both
        the setting name and the setting value.

    .. warning::

        From time to time Microsoft updates the Administrative templates on the
        machine. This can cause the policy name to change or the list of
        settings that must be applied at the same time. These settings often
        change between versions of Windows as well. For example, Windows Server
        2019 allows you to also specify a specific week of the month to apply
        the update.

    Another thing note is the long policy name returned by the `get` function:

    .. code-block:: bash

        Windows Components\Windows Update\Configure Automatic Updates:

    When we wrote the state for this policy we only used the final portion of
    the policy name, `Configure Automatic Updates`. This usually works fine, but
    if you are having problems, you may try the long policy name.

    When writing the long name in a state file either wrap the name in single
    quotes to make yaml see it as raw data, or escape the back slashes.

    .. code-block:: bash

        'Windows Components\Windows Update\Configure Automatic Updates:'

        or

        Windows Components\\Windows Update\\Configure Automatic Updates:
"""
import logging

import salt.utils.data
import salt.utils.dictdiffer
import salt.utils.json
import salt.utils.stringutils
import salt.utils.versions
import salt.utils.win_functions

log = logging.getLogger(__name__)
__virtualname__ = "lgpo"
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    load this state if the win_lgpo module exists
    """
    if "lgpo.set" in __salt__:
        return __virtualname__
    return False, "lgpo module could not be loaded"


def _compare_policies(new_policy, current_policy):
    """
    Helper function that returns ``True`` if the policies are the same,
    otherwise ``False``
    """
    # Compared dicts, lists, and strings
    if isinstance(new_policy, (str, int)):
        return new_policy == current_policy
    elif isinstance(new_policy, list):
        if isinstance(current_policy, list):
            return salt.utils.data.compare_lists(new_policy, current_policy) == {}
        else:
            return False
    elif isinstance(new_policy, dict):
        if isinstance(current_policy, dict):
            return salt.utils.data.compare_dicts(new_policy, current_policy) == {}
        else:
            return False


def _convert_to_unicode(data):
    """
    Helper function that makes sure all items in the dictionary are unicode for
    comparing the existing state with the desired state. This function is only
    needed for Python 2 and can be removed once we've migrated to Python 3.

    The data returned by the current settings sometimes has a mix of unicode and
    string values (these don't matter in Py3). This causes the comparison to
    say it's not in the correct state even though it is. They basically compares
    apples to apples, etc.

    Also, in Python 2, the utf-16 encoded strings remain utf-16 encoded (each
    character separated by `/x00`) In Python 3 it returns a utf-8 string. This
    will just remove all the null bytes (`/x00`), again comparing apples to
    apples.
    """
    if isinstance(data, str):
        data = data.replace("\x00", "")
        return salt.utils.stringutils.to_unicode(data)
    elif isinstance(data, dict):
        return {_convert_to_unicode(k): _convert_to_unicode(v) for k, v in data.items()}
    elif isinstance(data, list):
        return list(_convert_to_unicode(v) for v in data)
    else:
        return data


def set_(
    name,
    setting=None,
    policy_class=None,
    computer_policy=None,
    user_policy=None,
    cumulative_rights_assignments=True,
    adml_language="en-US",
):
    """
    Ensure the specified policy is set.

    .. warning::
        The ``setting`` argument cannot be used in conjunction with the
        ``computer_policy`` or ``user_policy`` arguments

    Args:
        name (str): The name of a single policy to configure

        setting (str, dict, list):
            The configuration setting for the single named policy. If this
            argument is used the ``computer_policy`` / ``user_policy`` arguments
            will be ignored

        policy_class (str):
            The policy class of the single named policy to configure. This can
            ``machine``, ``user``, or ``both``

        computer_policy (dict):
            A dictionary of containing the policy name and key/value pairs of a
            set of computer policies to configure. If this argument is used, the
            ``name`` / ``policy_class`` arguments will be ignored

        user_policy (dict):
            A dictionary of containing the policy name and key/value pairs of a
            set of user policies to configure. If this argument is used, the
            ``name`` / ``policy_class`` arguments will be ignored

        cumulative_rights_assignments (bool):
            If user rights assignments are being configured, determines if any
            user right assignment policies specified will be cumulative or
            explicit

        adml_language (str):
            The adml language to use for AMDX policy data/display conversions.
            Default is ``en-US``
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}
    policy_classes = ["machine", "computer", "user", "both"]
    class_map = {
        "computer": "Computer Configuration",
        "machine": "Computer Configuration",
        "user": "User Configuration",
    }
    if not setting and not computer_policy and not user_policy:
        msg = (
            "At least one of the parameters setting, computer_policy, or "
            "user_policy must be specified."
        )
        ret["result"] = False
        ret["comment"] = msg
        return ret
    if setting and not policy_class:
        msg = (
            "A single policy setting was specified but the policy_class "
            "was not specified."
        )
        ret["result"] = False
        ret["comment"] = msg
        return ret
    if setting and (computer_policy or user_policy):
        msg = (
            "The setting and computer_policy/user_policy parameters are "
            "mutually exclusive.  Please specify either a policy name and "
            "setting or a computer_policy and/or user_policy dict"
        )
        ret["result"] = False
        ret["comment"] = msg
        return ret
    if policy_class and policy_class.lower() not in policy_classes:
        msg = "The policy_class parameter must be one of the following: {}"
        ret["result"] = False
        ret["comment"] = msg
        return ret
    if not setting:
        if computer_policy and not isinstance(computer_policy, dict):
            msg = "The computer_policy must be specified as a dict."
            ret["result"] = False
            ret["comment"] = msg
            return ret
        if user_policy and not isinstance(user_policy, dict):
            msg = "The user_policy must be specified as a dict."
            ret["result"] = False
            ret["comment"] = msg
            return ret
    else:
        user_policy = {}
        computer_policy = {}
        if policy_class.lower() == "both":
            user_policy[name] = setting
            computer_policy[name] = setting
        elif policy_class.lower() == "user":
            user_policy[name] = setting
        elif policy_class.lower() in ["machine", "computer"]:
            computer_policy[name] = setting
    pol_data = {
        "user": {"requested_policy": user_policy, "policy_lookup": {}},
        "machine": {"requested_policy": computer_policy, "policy_lookup": {}},
    }

    current_policy = {}
    deprecation_comments = []
    for p_class, p_data in pol_data.items():
        if p_data["requested_policy"]:
            for p_name, _ in p_data["requested_policy"].items():
                lookup = __salt__["lgpo.get_policy_info"](
                    policy_name=p_name,
                    policy_class=p_class,
                    adml_language=adml_language,
                )
                if lookup["policy_found"]:
                    pol_data[p_class]["policy_lookup"][p_name] = lookup
                    # Since we found the policy, let's get the current setting
                    # as well
                    current_policy.setdefault(class_map[p_class], {})
                    current_policy[class_map[p_class]][p_name] = __salt__[
                        "lgpo.get_policy"
                    ](
                        policy_name=p_name,
                        policy_class=p_class,
                        adml_language=adml_language,
                        return_value_only=True,
                    )
                    # Validate element names
                    if isinstance(p_data["requested_policy"][p_name], dict):
                        valid_names = []
                        for element in lookup["policy_elements"]:
                            valid_names.extend(element["element_aliases"])
                        for e_name in p_data["requested_policy"][p_name]:
                            if e_name not in valid_names:
                                new_e_name = e_name.split(":")[-1].strip()
                                # If we find an invalid name, test the new
                                # format. If found, replace the old with the
                                # new
                                if new_e_name in valid_names:
                                    msg = (
                                        "The LGPO module changed the way "
                                        "it gets policy element names.\n"
                                        '"{}" is no longer valid.\n'
                                        'Please use "{}" instead.'
                                        "".format(e_name, new_e_name)
                                    )
                                    salt.utils.versions.warn_until("Phosphorus", msg)
                                    pol_data[p_class]["requested_policy"][p_name][
                                        new_e_name
                                    ] = pol_data[p_class]["requested_policy"][
                                        p_name
                                    ].pop(
                                        e_name
                                    )
                                    deprecation_comments.append(msg)
                                else:
                                    msg = "Invalid element name: {}".format(e_name)
                                    ret["comment"] = "\n".join(
                                        [ret["comment"], msg]
                                    ).strip()
                                    ret["result"] = False
                else:
                    ret["comment"] = "\n".join(
                        [ret["comment"], lookup["message"]]
                    ).strip()
                    ret["result"] = False
    if not ret["result"]:
        deprecation_comments.append(ret["comment"])
        ret["comment"] = "\n".join(deprecation_comments).strip()
        return ret

    log.debug("pol_data == %s", pol_data)
    log.debug("current policy == %s", current_policy)

    # compare policies
    policy_changes = []
    for p_class, p_data in pol_data.items():
        requested_policy = p_data.get("requested_policy")
        if requested_policy:
            for p_name, p_setting in requested_policy.items():
                if p_name in current_policy[class_map[p_class]]:
                    # compare the requested and current policies
                    log.debug(
                        "need to compare %s from current/requested policy", p_name
                    )

                    # resolve user names in the requested policy and the current
                    # policy so that we are comparing apples to apples
                    if p_data["policy_lookup"][p_name]["rights_assignment"]:
                        resolved_names = []
                        for name in p_data["requested_policy"][p_name]:
                            resolved_names.append(
                                salt.utils.win_functions.get_sam_name(name)
                            )
                        p_data["requested_policy"][p_name] = resolved_names
                        resolved_names = []
                        for name in current_policy[class_map[p_class]][p_name]:
                            resolved_names.append(
                                salt.utils.win_functions.get_sam_name(name)
                            )
                        current_policy[class_map[p_class]][p_name] = resolved_names

                    changes = False
                    requested_policy_json = salt.utils.json.dumps(
                        p_data["requested_policy"][p_name], sort_keys=True
                    ).lower()
                    current_policy_json = salt.utils.json.dumps(
                        current_policy[class_map[p_class]][p_name], sort_keys=True
                    ).lower()

                    requested_policy_check = salt.utils.json.loads(
                        requested_policy_json
                    )
                    current_policy_check = salt.utils.json.loads(current_policy_json)

                    # Are the requested and current policies identical
                    policies_are_equal = _compare_policies(
                        requested_policy_check, current_policy_check
                    )

                    if not policies_are_equal:
                        if (
                            p_data["policy_lookup"][p_name]["rights_assignment"]
                            and cumulative_rights_assignments
                        ):
                            for user in p_data["requested_policy"][p_name]:
                                if (
                                    user
                                    not in current_policy[class_map[p_class]][p_name]
                                ):
                                    user = salt.utils.win_functions.get_sam_name(user)
                                    if (
                                        user
                                        not in current_policy[class_map[p_class]][
                                            p_name
                                        ]
                                    ):
                                        changes = True
                        else:
                            changes = True
                        if changes:
                            log.debug("%s current policy != requested policy", p_name)
                            log.debug(
                                "We compared %s to %s",
                                requested_policy_json,
                                current_policy_json,
                            )
                            policy_changes.append(p_name)
                    else:
                        msg = '"{}" is already set'.format(p_name)
                        log.debug(msg)
                else:
                    policy_changes.append(p_name)
                    log.debug("policy %s is not set, we will configure it", p_name)
    if __opts__["test"]:
        if policy_changes:
            msg = "The following policies are set to change:\n{}".format(
                "\n".join(policy_changes)
            )
            ret["result"] = None
        else:
            msg = "All specified policies are properly configured"
        deprecation_comments.append(msg)
        ret["comment"] = "\n".join(deprecation_comments).strip()
    else:
        if policy_changes:
            _ret = __salt__["lgpo.set"](
                computer_policy=pol_data["machine"]["requested_policy"],
                user_policy=pol_data["user"]["requested_policy"],
                cumulative_rights_assignments=cumulative_rights_assignments,
                adml_language=adml_language,
            )
            if _ret:
                ret["result"] = _ret
                new_policy = {}
                for p_class, p_data in pol_data.items():
                    if p_data["requested_policy"]:
                        for p_name, p_setting in p_data["requested_policy"].items():
                            new_policy.setdefault(class_map[p_class], {})
                            new_policy[class_map[p_class]][p_name] = __salt__[
                                "lgpo.get_policy"
                            ](
                                policy_name=p_name,
                                policy_class=p_class,
                                adml_language=adml_language,
                                return_value_only=True,
                            )
                ret["changes"] = salt.utils.dictdiffer.deep_diff(
                    old=current_policy, new=new_policy
                )
                if ret["changes"]:
                    msg = "The following policies changed:\n{}".format(
                        "\n".join(policy_changes)
                    )
                else:
                    msg = "The following policies are in the correct state:\n{}".format(
                        "\n".join(policy_changes)
                    )
            else:
                msg = (
                    "Errors occurred while attempting to configure policies: {}".format(
                        _ret
                    )
                )
                ret["result"] = False
            deprecation_comments.append(msg)
            ret["comment"] = "\n".join(deprecation_comments).strip()
        else:
            msg = "All specified policies are properly configured"
            deprecation_comments.append(msg)
            ret["comment"] = "\n".join(deprecation_comments).strip()

    return ret



"""
Integration tests for the saltutil module.
"""


import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(autouse=True)
def refresh_pillar(salt_call_cli, salt_minion, salt_sub_minion):
    whitelist = {
        "modules": [],
    }
    ret = salt_call_cli.run("saltutil.sync_all", extmod_whitelist=whitelist)
    assert ret.exitcode == 0
    assert ret.json
    try:
        yield
    finally:
        ret = salt_call_cli.run("saltutil.sync_all")
        assert ret.exitcode == 0
        assert ret.json


@pytest.mark.slow_test
def test_sync_all(salt_call_cli):
    """
    Test syncing all ModuleCase
    """
    expected_return = {
        "engines": [],
        "clouds": [],
        "grains": [],
        "beacons": [],
        "utils": [],
        "returners": [],
        "modules": [
            "modules.depends_versioned",
            "modules.depends_versionless",
            "modules.mantest",
            "modules.override_test",
            "modules.runtests_decorators",
            "modules.runtests_helpers",
            "modules.salttest",
        ],
        "renderers": [],
        "log_handlers": [],
        "matchers": [],
        "states": [],
        "sdb": [],
        "proxymodules": [],
        "executors": [],
        "output": [],
        "thorium": [],
        "serializers": [],
    }
    ret = salt_call_cli.run("saltutil.sync_all")
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == expected_return


@pytest.mark.slow_test
def test_sync_all_whitelist(salt_call_cli):
    """
    Test syncing all ModuleCase with whitelist
    """
    expected_return = {
        "engines": [],
        "clouds": [],
        "grains": [],
        "beacons": [],
        "utils": [],
        "returners": [],
        "modules": ["modules.salttest"],
        "renderers": [],
        "log_handlers": [],
        "matchers": [],
        "states": [],
        "sdb": [],
        "proxymodules": [],
        "executors": [],
        "output": [],
        "thorium": [],
        "serializers": [],
    }
    ret = salt_call_cli.run(
        "saltutil.sync_all", extmod_whitelist={"modules": ["salttest"]}
    )
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == expected_return


@pytest.mark.slow_test
def test_sync_all_blacklist(salt_call_cli):
    """
    Test syncing all ModuleCase with blacklist
    """
    expected_return = {
        "engines": [],
        "clouds": [],
        "grains": [],
        "beacons": [],
        "utils": [],
        "returners": [],
        "modules": [
            "modules.mantest",
            "modules.override_test",
            "modules.runtests_helpers",
            "modules.salttest",
        ],
        "renderers": [],
        "log_handlers": [],
        "matchers": [],
        "states": [],
        "sdb": [],
        "proxymodules": [],
        "executors": [],
        "output": [],
        "thorium": [],
        "serializers": [],
    }
    ret = salt_call_cli.run(
        "saltutil.sync_all",
        extmod_blacklist={
            "modules": [
                "runtests_decorators",
                "depends_versioned",
                "depends_versionless",
            ]
        },
    )
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == expected_return


@pytest.mark.slow_test
def test_sync_all_blacklist_and_whitelist(salt_call_cli):
    """
    Test syncing all ModuleCase with whitelist and blacklist
    """
    expected_return = {
        "engines": [],
        "clouds": [],
        "grains": [],
        "beacons": [],
        "utils": [],
        "returners": [],
        "executors": [],
        "modules": [],
        "renderers": [],
        "log_handlers": [],
        "matchers": [],
        "states": [],
        "sdb": [],
        "proxymodules": [],
        "output": [],
        "thorium": [],
        "serializers": [],
    }
    ret = salt_call_cli.run(
        "saltutil.sync_all",
        extmod_whitelist={"modules": ["runtests_decorators"]},
        extmod_blacklist={"modules": ["runtests_decorators"]},
    )
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == expected_return



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.win_autoruns as win_autoruns
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

KEY = [
    "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
    "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run /reg:64",
    "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
]


class WinAutorunsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_autoruns
    """

    def setup_loader_modules(self):
        return {win_autoruns: {}}

    # 'list_' function tests: 1

    def test_list(self):
        """
        Test if it enables win_autoruns the service on the server
        """
        with patch("os.listdir", MagicMock(return_value=[])):
            ret = {KEY[0]: ["SALT"], KEY[1]: ["SALT"], KEY[2]: ["SALT"]}
            mock = MagicMock(return_value="Windows 7")
            with patch.dict(win_autoruns.__grains__, {"osfullname": mock}):
                mock = MagicMock(return_value="SALT")
                with patch.dict(win_autoruns.__salt__, {"cmd.run": mock}):
                    self.assertDictEqual(win_autoruns.list_(), ret)



#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A non-blocking, single-threaded HTTP server.

Typical applications have little direct interaction with the `HTTPServer`
class except to start a server at the beginning of the process
(and even that is often done indirectly via `tornado.web.Application.listen`).

.. versionchanged:: 4.0

   The ``HTTPRequest`` class that used to live in this module has been moved
   to `tornado.httputil.HTTPServerRequest`.  The old name remains as an alias.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import socket

from salt.ext.tornado.escape import native_str
from salt.ext.tornado.http1connection import HTTP1ServerConnection, HTTP1ConnectionParameters
from salt.ext.tornado import gen
from salt.ext.tornado import httputil
from salt.ext.tornado import iostream
from salt.ext.tornado import netutil
from salt.ext.tornado.tcpserver import TCPServer
from salt.ext.tornado.util import Configurable


class HTTPServer(TCPServer, Configurable,
                 httputil.HTTPServerConnectionDelegate):
    r"""A non-blocking, single-threaded HTTP server.

    A server is defined by a subclass of `.HTTPServerConnectionDelegate`,
    or, for backwards compatibility, a callback that takes an
    `.HTTPServerRequest` as an argument. The delegate is usually a
    `tornado.web.Application`.

    `HTTPServer` supports keep-alive connections by default
    (automatically for HTTP/1.1, or for HTTP/1.0 when the client
    requests ``Connection: keep-alive``).

    If ``xheaders`` is ``True``, we support the
    ``X-Real-Ip``/``X-Forwarded-For`` and
    ``X-Scheme``/``X-Forwarded-Proto`` headers, which override the
    remote IP and URI scheme/protocol for all requests.  These headers
    are useful when running Tornado behind a reverse proxy or load
    balancer.  The ``protocol`` argument can also be set to ``https``
    if Tornado is run behind an SSL-decoding proxy that does not set one of
    the supported ``xheaders``.

    By default, when parsing the ``X-Forwarded-For`` header, Tornado will
    select the last (i.e., the closest) address on the list of hosts as the
    remote host IP address.  To select the next server in the chain, a list of
    trusted downstream hosts may be passed as the ``trusted_downstream``
    argument.  These hosts will be skipped when parsing the ``X-Forwarded-For``
    header.

    To make this server serve SSL traffic, send the ``ssl_options`` keyword
    argument with an `ssl.SSLContext` object. For compatibility with older
    versions of Python ``ssl_options`` may also be a dictionary of keyword
    arguments for the `ssl.wrap_socket` method.::

       ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
       ssl_ctx.load_cert_chain(os.path.join(data_dir, "mydomain.crt"),
                               os.path.join(data_dir, "mydomain.key"))
       HTTPServer(applicaton, ssl_options=ssl_ctx)

    `HTTPServer` initialization follows one of three patterns (the
    initialization methods are defined on `tornado.tcpserver.TCPServer`):

    1. `~tornado.tcpserver.TCPServer.listen`: simple single-process::

            server = HTTPServer(app)
            server.listen(8888)
            IOLoop.current().start()

       In many cases, `tornado.web.Application.listen` can be used to avoid
       the need to explicitly create the `HTTPServer`.

    2. `~tornado.tcpserver.TCPServer.bind`/`~tornado.tcpserver.TCPServer.start`:
       simple multi-process::

            server = HTTPServer(app)
            server.bind(8888)
            server.start(0)  # Forks multiple sub-processes
            IOLoop.current().start()

       When using this interface, an `.IOLoop` must *not* be passed
       to the `HTTPServer` constructor.  `~.TCPServer.start` will always start
       the server on the default singleton `.IOLoop`.

    3. `~tornado.tcpserver.TCPServer.add_sockets`: advanced multi-process::

            sockets = tornado.netutil.bind_sockets(8888)
            tornado.process.fork_processes(0)
            server = HTTPServer(app)
            server.add_sockets(sockets)
            IOLoop.current().start()

       The `~.TCPServer.add_sockets` interface is more complicated,
       but it can be used with `tornado.process.fork_processes` to
       give you more flexibility in when the fork happens.
       `~.TCPServer.add_sockets` can also be used in single-process
       servers if you want to create your listening sockets in some
       way other than `tornado.netutil.bind_sockets`.

    .. versionchanged:: 4.0
       Added ``decompress_request``, ``chunk_size``, ``max_header_size``,
       ``idle_connection_timeout``, ``body_timeout``, ``max_body_size``
       arguments.  Added support for `.HTTPServerConnectionDelegate`
       instances as ``request_callback``.

    .. versionchanged:: 4.1
       `.HTTPServerConnectionDelegate.start_request` is now called with
       two arguments ``(server_conn, request_conn)`` (in accordance with the
       documentation) instead of one ``(request_conn)``.

    .. versionchanged:: 4.2
       `HTTPServer` is now a subclass of `tornado.util.Configurable`.

    .. versionchanged:: 4.5
       Added the ``trusted_downstream`` argument.
    """
    def __init__(self, *args, **kwargs):
        # Ignore args to __init__; real initialization belongs in
        # initialize since we're Configurable. (there's something
        # weird in initialization order between this class,
        # Configurable, and TCPServer so we can't leave __init__ out
        # completely)
        pass

    def initialize(self, request_callback, no_keep_alive=False, io_loop=None,
                   xheaders=False, ssl_options=None, protocol=None,
                   decompress_request=False,
                   chunk_size=None, max_header_size=None,
                   idle_connection_timeout=None, body_timeout=None,
                   max_body_size=None, max_buffer_size=None,
                   trusted_downstream=None):
        self.request_callback = request_callback
        self.no_keep_alive = no_keep_alive
        self.xheaders = xheaders
        self.protocol = protocol
        self.conn_params = HTTP1ConnectionParameters(
            decompress=decompress_request,
            chunk_size=chunk_size,
            max_header_size=max_header_size,
            header_timeout=idle_connection_timeout or 3600,
            max_body_size=max_body_size,
            body_timeout=body_timeout,
            no_keep_alive=no_keep_alive)
        TCPServer.__init__(self, io_loop=io_loop, ssl_options=ssl_options,
                           max_buffer_size=max_buffer_size,
                           read_chunk_size=chunk_size)
        self._connections = set()
        self.trusted_downstream = trusted_downstream

    @classmethod
    def configurable_base(cls):
        return HTTPServer

    @classmethod
    def configurable_default(cls):
        return HTTPServer

    @gen.coroutine
    def close_all_connections(self):
        while self._connections:
            # Peek at an arbitrary element of the set
            conn = next(iter(self._connections))
            yield conn.close()

    def handle_stream(self, stream, address):
        context = _HTTPRequestContext(stream, address,
                                      self.protocol,
                                      self.trusted_downstream)
        conn = HTTP1ServerConnection(
            stream, self.conn_params, context)
        self._connections.add(conn)
        conn.start_serving(self)

    def start_request(self, server_conn, request_conn):
        if isinstance(self.request_callback, httputil.HTTPServerConnectionDelegate):
            delegate = self.request_callback.start_request(server_conn, request_conn)
        else:
            delegate = _CallableAdapter(self.request_callback, request_conn)

        if self.xheaders:
            delegate = _ProxyAdapter(delegate, request_conn)

        return delegate

    def on_close(self, server_conn):
        self._connections.remove(server_conn)


class _CallableAdapter(httputil.HTTPMessageDelegate):
    def __init__(self, request_callback, request_conn):
        self.connection = request_conn
        self.request_callback = request_callback
        self.request = None
        self.delegate = None
        self._chunks = []

    def headers_received(self, start_line, headers):
        self.request = httputil.HTTPServerRequest(
            connection=self.connection, start_line=start_line,
            headers=headers)

    def data_received(self, chunk):
        self._chunks.append(chunk)

    def finish(self):
        self.request.body = b''.join(self._chunks)
        self.request._parse_body()
        self.request_callback(self.request)

    def on_connection_close(self):
        self._chunks = None


class _HTTPRequestContext(object):
    def __init__(self, stream, address, protocol, trusted_downstream=None):
        self.address = address
        # Save the socket's address family now so we know how to
        # interpret self.address even after the stream is closed
        # and its socket attribute replaced with None.
        if stream.socket is not None:
            self.address_family = stream.socket.family
        else:
            self.address_family = None
        # In HTTPServerRequest we want an IP, not a full socket address.
        if (self.address_family in (socket.AF_INET, socket.AF_INET6) and
                address is not None):
            self.remote_ip = address[0]
        else:
            # Unix (or other) socket; fake the remote address.
            self.remote_ip = '0.0.0.0'
        if protocol:
            self.protocol = protocol
        elif isinstance(stream, iostream.SSLIOStream):
            self.protocol = "https"
        else:
            self.protocol = "http"
        self._orig_remote_ip = self.remote_ip
        self._orig_protocol = self.protocol
        self.trusted_downstream = set(trusted_downstream or [])

    def __str__(self):
        if self.address_family in (socket.AF_INET, socket.AF_INET6):
            return self.remote_ip
        elif isinstance(self.address, bytes):
            # Python 3 with the -bb option warns about str(bytes),
            # so convert it explicitly.
            # Unix socket addresses are str on mac but bytes on linux.
            return native_str(self.address)
        else:
            return str(self.address)

    def _apply_xheaders(self, headers):
        """Rewrite the ``remote_ip`` and ``protocol`` fields."""
        # Squid uses X-Forwarded-For, others use X-Real-Ip
        ip = headers.get("X-Forwarded-For", self.remote_ip)
        # Skip trusted downstream hosts in X-Forwarded-For list
        for ip in (cand.strip() for cand in reversed(ip.split(','))):
            if ip not in self.trusted_downstream:
                break
        ip = headers.get("X-Real-Ip", ip)
        if netutil.is_valid_ip(ip):
            self.remote_ip = ip
        # AWS uses X-Forwarded-Proto
        proto_header = headers.get(
            "X-Scheme", headers.get("X-Forwarded-Proto",
                                    self.protocol))
        if proto_header in ("http", "https"):
            self.protocol = proto_header

    def _unapply_xheaders(self):
        """Undo changes from `_apply_xheaders`.

        Xheaders are per-request so they should not leak to the next
        request on the same connection.
        """
        self.remote_ip = self._orig_remote_ip
        self.protocol = self._orig_protocol


class _ProxyAdapter(httputil.HTTPMessageDelegate):
    def __init__(self, delegate, request_conn):
        self.connection = request_conn
        self.delegate = delegate

    def headers_received(self, start_line, headers):
        self.connection.context._apply_xheaders(headers)
        return self.delegate.headers_received(start_line, headers)

    def data_received(self, chunk):
        return self.delegate.data_received(chunk)

    def finish(self):
        self.delegate.finish()
        self._cleanup()

    def on_connection_close(self):
        self.delegate.on_connection_close()
        self._cleanup()

    def _cleanup(self):
        self.connection.context._unapply_xheaders()


HTTPRequest = httputil.HTTPServerRequest



"""
    :codeauthor: Marek Skrobacki <skrobul@skrobul.com>
"""


import os

import salt.modules.s6 as s6
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class S6TestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.s6
    """

    def setup_loader_modules(self):
        return {s6: {"SERVICE_DIR": "/etc/service"}}

    # 'start' function tests: 1

    def test_start(self):
        """
        Test if it starts service via s6-svc.
        """
        mock_ret = MagicMock(return_value=False)
        with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
            self.assertTrue(s6.start("ssh"))

    # 'stop' function tests: 1

    def test_stop(self):
        """
        Test if it stops service via s6.
        """
        mock_ret = MagicMock(return_value=False)
        with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
            self.assertTrue(s6.stop("ssh"))

    # 'term' function tests: 1

    def test_term(self):
        """
        Test if it send a TERM to service via s6.
        """
        mock_ret = MagicMock(return_value=False)
        with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
            self.assertTrue(s6.term("ssh"))

    # 'reload_' function tests: 1

    def test_reload(self):
        """
        Test if it send a HUP to service via s6.
        """
        mock_ret = MagicMock(return_value=False)
        with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
            self.assertTrue(s6.reload_("ssh"))

    # 'restart' function tests: 1

    def test_restart(self):
        """
        Test if it restart service via s6. This will stop/start service.
        """
        mock_ret = MagicMock(return_value=False)
        with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
            self.assertTrue(s6.restart("ssh"))

    # 'full_restart' function tests: 1

    def test_full_restart(self):
        """
        Test if it calls s6.restart() function.
        """
        mock_ret = MagicMock(return_value=False)
        with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
            self.assertIsNone(s6.full_restart("ssh"))

    # 'status' function tests: 1

    def test_status(self):
        """
        Test if it return the status for a service via s6,
        return pid if running.
        """
        mock_run = MagicMock(return_value="salt")
        with patch.dict(s6.__salt__, {"cmd.run_stdout": mock_run}):
            self.assertEqual(s6.status("ssh"), "")

    # 'available' function tests: 1

    def test_available(self):
        """
        Test if it returns ``True`` if the specified service is available,
        otherwise returns ``False``.
        """
        with patch.object(os, "listdir", MagicMock(return_value=["/etc/service"])):
            self.assertTrue(s6.available("/etc/service"))

    # 'missing' function tests: 1

    def test_missing(self):
        """
        Test if it returns ``True`` if the specified service is not available,
        otherwise returns ``False``.
        """
        with patch.object(os, "listdir", MagicMock(return_value=["/etc/service"])):
            self.assertTrue(s6.missing("foo"))

    # 'get_all' function tests: 1

    def test_get_all(self):
        """
        Test if it return a list of all available services.
        """
        with patch.object(os, "listdir", MagicMock(return_value=["/etc/service"])):
            self.assertListEqual(s6.get_all(), ["/etc/service"])



"""
    :codeauthor: Megan Wilhite<mwilhite@saltstack.com>
"""


import pytest
import salt.modules.mac_service as mac_service
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MacServiceTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.mac_service module
    """

    def setup_loader_modules(self):
        return {mac_service: {"__context__": {}}}

    def test_service_disabled_when_enabled(self):
        """
        test service.disabled when service is enabled
        """
        srv_name = "com.apple.atrun"
        cmd = (
            'disabled services = {\n\t"com.saltstack.salt.minion" =>'
            ' false\n\t"com.apple.atrun" => false\n{'
        )
        domain_ret = MagicMock(return_value=("", ""))
        with patch.object(mac_service, "_get_domain_target", domain_ret):
            with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
                assert mac_service.disabled(srv_name) is False

    def test_service_disabled_when_disabled(self):
        """
        test service.disabled when service is disabled
        """
        srv_name = "com.apple.atrun"
        cmd = (
            'disabled services = {\n\t"com.saltstack.salt.minion" =>'
            ' false\n\t"com.apple.atrun" => true\n{'
        )
        domain_ret = MagicMock(return_value=("", ""))
        with patch.object(mac_service, "_get_domain_target", domain_ret):
            with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
                assert mac_service.disabled(srv_name) is True

    def test_service_disabled_srvname_wrong(self):
        """
        test service.disabled when service is just slightly wrong
        """
        srv_names = ["com.apple.atru", "com", "apple"]
        cmd = (
            'disabled services = {\n\t"com.saltstack.salt.minion" =>'
            ' false\n\t"com.apple.atrun" => true\n}'
        )
        domain_ret = MagicMock(return_value=("", ""))
        with patch.object(mac_service, "_get_domain_target", domain_ret):
            for name in srv_names:
                with patch.object(
                    mac_service, "launchctl", MagicMock(return_value=cmd)
                ):
                    assert mac_service.disabled(name) is False

    def test_service_disabled_status_upper_case(self):
        """
        test service.disabled when disabled status is uppercase
        """
        srv_name = "com.apple.atrun"
        cmd = (
            'disabled services = {\n\t"com.saltstack.salt.minion" =>'
            ' false\n\t"com.apple.atrun" => True\n{'
        )
        domain_ret = MagicMock(return_value=("", ""))
        with patch.object(mac_service, "_get_domain_target", domain_ret):
            with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
                assert mac_service.disabled(srv_name) is True

    def test_service_enabled_when_enabled(self):
        """
        test service.enabled when not disabled
        """
        mock_cmd = MagicMock(return_value=False)
        with patch.dict(mac_service.__salt__, {"service.disabled": mock_cmd}):
            assert mac_service.enabled("com.apple.atrun") is True

    def test_service_enabled_when_disabled(self):
        """
        test service.enabled if service is disabled
        """
        mock_cmd = MagicMock(return_value=True)
        with patch.dict(mac_service.__salt__, {"service.disabled": mock_cmd}):
            assert mac_service.enabled("com.apple.atrun") is False

    def test_service_loaded_when_true(self):
        """
        test service.loaded with a loaded service.
        """
        mock_cmd = MagicMock(return_value="some_service_string")
        with patch.dict(mac_service.__salt__, {"service.list": mock_cmd}):
            assert mac_service.loaded("com.apple.atrun") is True

    def test_service_loaded_when_false(self):
        """
        test service.loaded with an unloaded service.
        """
        mock_cmd = MagicMock(side_effect=CommandExecutionError)
        with patch.dict(mac_service.__salt__, {"service.list": mock_cmd}):
            assert mac_service.loaded("com.apple.atrun") is False

    def test_service_keep_alive_pathstate_file_rm(self):
        """
        test _always_running_service when keep_alive
        has pathstate set in plist file and file doesn't exist
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": {"PathState": {"/private/etc/ntp.conf": True}},
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            with patch("os.path.exists", MagicMock(return_value=False)):
                assert mac_service._always_running_service(srv_name) is False

    def test_service_keep_alive_empty(self):
        """
        test _always_running_service when keep_alive
        is empty
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": {},
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            with patch("os.path.exists", MagicMock(return_value=False)):
                assert mac_service._always_running_service(srv_name) is False

    def test_service_keep_alive_pathstate_false(self):
        """
        test _always_running_service when keep_alive
        has pathstate set in plist file and file is false
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": {"PathState": {"/private/etc/ntp.conf": False}},
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            with patch("os.path.exists", MagicMock(return_value=False)):
                assert mac_service._always_running_service(srv_name) is True

    def test_service_keep_alive_pathstate(self):
        """
        test _always_running_service when keep_alive
        has pathstate set in plist file
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": {"PathState": {"/private/etc/ntp.conf": True}},
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            with patch("os.path.exists", MagicMock(return_value=True)):
                assert mac_service._always_running_service(srv_name) is True

    def test_service_keep_alive(self):
        """
        test _always_running_service when keep_alive set
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": True,
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            assert mac_service._always_running_service(srv_name) is True

    def test_service_keep_alive_false(self):
        """
        test _always_running_service when keep_alive False
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": False,
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            assert mac_service._always_running_service(srv_name) is False

    def test_service_keep_alive_missing(self):
        """
        test _always_running_service when keep_alive not in dict
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            assert mac_service._always_running_service(srv_name) is False

    def test_service_keep_alive_wrong_setting(self):
        """
        test _always_running_service when keep_alive
        has pathstate set in plist file
        """
        srv_name = "com.apple.atrun"
        info = {
            "plist": {
                "EnableTransactions": True,
                "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
                "Label": "org.ntp.ntpd",
                "KeepAlive": {"Doesnotexist": {"doesnt_exist": True}},
            }
        }

        with patch.object(mac_service, "show", MagicMock(return_value=info)):
            assert mac_service._always_running_service(srv_name) is False

    def test_service_name_change_salt_minion(self):
        srv_name = "salt-minion"
        info = {
            "com.saltstack.salt.minion": {
                "file_name": "com.saltstack.salt.minion.plist",
                "file_path": "/Library/LaunchDaemons/com.saltstack.salt.minion.plist",
                "plist": {
                    "HardResourceLimits": {"NumberOfFiles": 100000},
                    "KeepAlive": True,
                    "Label": "com.saltstack.salt.minion",
                    "ProgramArguments": ["/opt/salt/bin/start-salt-minion.sh"],
                    "RunAtLoad": True,
                    "SoftResourceLimits": {"NumberOfFiles": 100000},
                },
            }
        }
        with patch.dict(
            mac_service.__utils__,
            {"mac_utils.available_services": MagicMock(return_value=info)},
        ):
            assert (
                mac_service._get_service(srv_name) == info["com.saltstack.salt.minion"]
            )

    def test_service_name_change_salt_master(self):
        srv_name = "salt-master"
        info = {
            "com.saltstack.salt.master": {
                "file_name": "com.saltstack.salt.master.plist",
                "file_path": "/Library/LaunchDaemons/com.saltstack.salt.master.plist",
                "plist": {
                    "HardResourceLimits": {"NumberOfFiles": 100000},
                    "KeepAlive": True,
                    "Label": "com.saltstack.salt.master",
                    "ProgramArguments": ["/opt/salt/bin/start-salt-master.sh"],
                    "RunAtLoad": True,
                    "SoftResourceLimits": {"NumberOfFiles": 100000},
                },
            }
        }
        with patch.dict(
            mac_service.__utils__,
            {"mac_utils.available_services": MagicMock(return_value=info)},
        ):
            assert (
                mac_service._get_service(srv_name) == info["com.saltstack.salt.master"]
            )

    def test_service_name_change_salt_api(self):
        srv_name = "salt-api"
        info = {
            "com.saltstack.salt.api": {
                "file_name": "com.saltstack.salt.api.plist",
                "file_path": "/Library/LaunchDaemons/com.saltstack.salt.api.plist",
                "plist": {
                    "HardResourceLimits": {"NumberOfFiles": 100000},
                    "KeepAlive": True,
                    "Label": "com.saltstack.salt.api",
                    "ProgramArguments": ["/opt/salt/bin/start-salt-api.sh"],
                    "RunAtLoad": True,
                    "SoftResourceLimits": {"NumberOfFiles": 100000},
                },
            }
        }
        with patch.dict(
            mac_service.__utils__,
            {"mac_utils.available_services": MagicMock(return_value=info)},
        ):
            assert mac_service._get_service(srv_name) == info["com.saltstack.salt.api"]

    def test_service_name_change_salt_syndic(self):
        srv_name = "salt-syndic"
        info = {
            "com.saltstack.salt.syndic": {
                "file_name": "com.saltstack.salt.syndic.plist",
                "file_path": "/Library/LaunchDaemons/com.saltstack.salt.syndic.plist",
                "plist": {
                    "HardResourceLimits": {"NumberOfFiles": 100000},
                    "KeepAlive": True,
                    "Label": "com.saltstack.salt.syndic",
                    "ProgramArguments": ["/opt/salt/bin/start-salt-syndic.sh"],
                    "RunAtLoad": True,
                    "SoftResourceLimits": {"NumberOfFiles": 100000},
                },
            }
        }
        with patch.dict(
            mac_service.__utils__,
            {"mac_utils.available_services": MagicMock(return_value=info)},
        ):
            assert (
                mac_service._get_service(srv_name) == info["com.saltstack.salt.syndic"]
            )

    def test_service_restart_already_loaded(self):
        mock_cmd = MagicMock(return_value=True)
        salt_dict = {
            "service.loaded": mock_cmd,
            "service.stop": mock_cmd,
            "service.start": mock_cmd,
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            assert mac_service.restart("com.salt") is True

    def test_service_restart_not_loaded(self):
        salt_dict = {
            "service.loaded": MagicMock(return_value=False),
            "service.start": MagicMock(return_value=True),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            assert mac_service.restart("com.salt") is True

    def test_service_restart_failed_stop(self):
        salt_dict = {
            "service.loaded": MagicMock(return_value=True),
            "service.stop": MagicMock(side_effect=CommandExecutionError),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            with pytest.raises(CommandExecutionError):
                assert mac_service.restart("com.salt")

    def test_service_restart_failed_start(self):
        salt_dict = {
            "service.loaded": MagicMock(return_value=False),
            "service.start": MagicMock(side_effect=CommandExecutionError),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            with pytest.raises(CommandExecutionError):
                assert mac_service.restart("com.salt")

    def test_service_status_no_service(self):
        """
        Test service status with no service found
        """
        with patch.object(
            mac_service, "_get_service", MagicMock(side_effect=CommandExecutionError)
        ):
            assert mac_service.status("com.salt") is False

    @patch.object(mac_service, "_launch_agent", lambda _: False)
    @patch.object(mac_service, "_get_service", lambda _: {"": ""})
    @patch.object(mac_service, "_always_running_service", lambda _: True)
    def test_service_status_on_daemon_with_pid(self):
        """
        Test service status on dameon with PID.
        """
        mock_service_list = (
            '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
            ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" = 0;\n\t"PID" ='
            ' 218;\n\t"Program" = "/opt/salt";\n\t\t"--disable-keepalive";\n\t);\n};'
        )
        salt_dict = {
            "service.list": MagicMock(return_value=mock_service_list),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            assert mac_service.status("com.salt") is True

    @patch.object(mac_service, "_launch_agent", lambda _: True)
    @patch.object(mac_service, "_get_service", lambda _: {"": ""})
    @patch.object(mac_service, "_always_running_service", lambda _: True)
    def test_service_status_on_agent_with_pid(self):
        """
        Test service status on LaunchAgent with PID.
        """
        mock_service_list = (
            '{\n\t"LimitLoadToSessionType" = "Aqua";\n\t"Label" ='
            ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" = 19968;\n\t"PID"'
            ' = 218;\n\t"Program" = "/opt/salt";\n\t"ProgramArguments" ='
            ' (\n\t\t"/opt/salt";\n\t\t"--syslog";\n\t);\n};'
        )
        salt_dict = {
            "service.list": MagicMock(return_value=mock_service_list),
        }
        utils_dict = {
            "mac_utils.console_user": MagicMock(return_value="spongebob"),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            with patch.dict(mac_service.__utils__, utils_dict):
                assert mac_service.status("com.salt") is True

    @patch.object(mac_service, "_launch_agent", lambda _: True)
    @patch.object(mac_service, "_get_service", lambda _: {"": ""})
    @patch.object(mac_service, "_always_running_service", lambda _: True)
    def test_service_status_on_agent_with_no_pid_and_should_be_running(self):
        """
        Test service status on LaunchAgent with No PID and should be running.
        """
        mock_service_list = (
            '{\n\t"LimitLoadToSessionType" = "Aqua";\n\t"Label" ='
            ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
            ' 19968;\n\t"Program" = "/opt/salt";\n\t"ProgramArguments" ='
            ' (\n\t\t"/opt/salt";\n\t\t"--syslog";\n\t);\n};'
        )
        salt_dict = {
            "service.list": MagicMock(return_value=mock_service_list),
        }
        utils_dict = {
            "mac_utils.console_user": MagicMock(return_value="spongebob"),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            with patch.dict(mac_service.__utils__, utils_dict):
                assert mac_service.status("com.salt") is False

    @patch.object(mac_service, "_launch_agent", lambda _: False)
    @patch.object(mac_service, "_get_service", lambda _: {"": ""})
    @patch.object(mac_service, "_always_running_service", lambda _: True)
    def test_service_status_on_daemon_with_no_pid_and_should_be_running(self):
        """
        Test service status on LaunchDaemon with no PID and an
        always running service that is loaded.
        """
        mock_service_list = (
            '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
            ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
            ' 19968;\n\t"Program" = "/opt/salt.sh";\n\t"ProgramArguments" ='
            ' (\n\t\t"/opt/salt.sh";\n\t\t"--disable-keepalive";\n\t);\n};'
        )
        salt_dict = {
            "service.list": MagicMock(return_value=mock_service_list),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            assert mac_service.status("com.salt") is False

    @patch.object(mac_service, "_launch_agent", lambda _: False)
    @patch.object(mac_service, "_get_service", lambda _: {"": ""})
    @patch.object(mac_service, "_always_running_service", lambda _: False)
    def test_service_status_on_daemon_with_no_pid_and_not_always_running(self):
        """
        Test service status on LaunchDaemon with no PID and not an always
        running service.
        """
        mock_service_list = (
            '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
            ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
            ' 19968;\n\t"Program" = "/opt/salt.sh";\n\t"ProgramArguments" ='
            ' (\n\t\t"/opt/salt.sh";\n\t\t"--disable-keepalive";\n\t);\n};'
        )
        salt_dict = {
            "service.list": MagicMock(return_value=mock_service_list),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            assert mac_service.status("com.salt") is True

    @patch.object(mac_service, "_launch_agent", lambda _: False)
    @patch.object(mac_service, "_get_service", lambda _: {"": ""})
    @patch.object(mac_service, "_always_running_service", lambda _: False)
    def test_service_status_on_daemon_with_failing_list_check(self):
        """
        Test service status on LaunchDaemon with no PID on an
        always running service that is loaded.
        """
        mock_service_list = (
            '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
            ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
            ' 19968;\n\t"Program" = "/opt/salt.sh";\n\t"ProgramArguments" ='
            ' (\n\t\t"/opt/salt.sh";\n\t\t"--disable-keepalive";\n\t);\n};'
        )
        salt_dict = {
            "service.list": MagicMock(side_effect=CommandExecutionError),
        }
        with patch.dict(mac_service.__salt__, salt_dict):
            assert mac_service.status("com.salt") is False

    def test_get_service_on_service_dead(self):
        """
        Test service.dead changes.
        https://github.com/saltstack/salt/issues/57907
        """
        utils_dict = {
            "mac_utils.available_services": MagicMock(return_value={}),
        }
        context_dict = {
            "using_cached_services": True,
            "service.state": "dead",
        }
        name_in_service = MagicMock(side_effect=[{}, {"com.salt": True}])
        with patch.dict(mac_service.__utils__, utils_dict):
            with patch.object(mac_service, "_name_in_services", name_in_service):
                with patch.dict(mac_service.__context__, context_dict):
                    with pytest.raises(CommandExecutionError):
                        assert mac_service._get_service("com.salt")
                # find the service on a second go with no service.dead
                with patch.dict(mac_service.__context__, {}):
                    assert mac_service._get_service("com.salt") == {"com.salt": True}



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.rabbitmq_policy as rabbitmq_policy
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rabbitmq_policy: {}}


# 'present' function tests: 1


def test_present():
    """
    Test to ensure the RabbitMQ policy exists.
    """
    name = "HA"
    pattern = ".*"
    definition = '{"ha-mode":"all"}'

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(
        side_effect=[
            {
                "/": {
                    name: {"pattern": pattern, "definition": definition, "priority": 0}
                }
            },
            {},
        ]
    )
    with patch.dict(rabbitmq_policy.__salt__, {"rabbitmq.list_policies": mock}):
        comt = "Policy / HA is already present"
        ret.update({"comment": comt})
        assert rabbitmq_policy.present(name, pattern, definition) == ret

        with patch.dict(rabbitmq_policy.__opts__, {"test": True}):
            comment = "Policy / HA is set to be created"
            changes = {"new": "HA", "old": {}}
            ret.update({"comment": comment, "result": None, "changes": changes})
            assert rabbitmq_policy.present(name, pattern, definition) == ret


# 'absent' function tests: 1


def test_absent():
    """
    Test to ensure the named policy is absent.
    """
    name = "HA"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(rabbitmq_policy.__salt__, {"rabbitmq.policy_exists": mock}):
        comment = "Policy '/ HA' is not present."
        ret.update({"comment": comment})
        assert rabbitmq_policy.absent(name) == ret

        with patch.dict(rabbitmq_policy.__opts__, {"test": True}):
            comment = "Policy '/ HA' will be removed."
            changes = {"new": "", "old": "HA"}
            ret.update({"comment": comment, "result": None, "changes": changes})
            assert rabbitmq_policy.absent(name) == ret



"""
    :codeauthor: Mike Place (mp@saltstack.com)


    tests.unit.returners.smtp_return_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import salt.returners.smtp_return as smtp
from salt.utils.jinja import SaltCacheLoader
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    import gnupg  # pylint: disable=unused-import

    HAS_GNUPG = True
except ImportError:
    HAS_GNUPG = False


class SMTPReturnerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test SMTP returner
    """

    def setup_loader_modules(self):
        return {
            smtp: {
                "__opts__": {
                    "extension_modules": "",
                    "optimization_order": [0, 1, 2],
                    "renderer": "jinja|yaml",
                    "renderer_blacklist": [],
                    "renderer_whitelist": [],
                    "file_roots": {},
                    "pillar_roots": {},
                    "cachedir": "/",
                    "master_uri": "tcp://127.0.0.1:4505",
                    "pki_dir": "/",
                    "keysize": 2048,
                    "id": "test",
                    "__role": "minion",
                }
            }
        }

    def _test_returner(self, mocked_smtplib):  # pylint: disable=unused-argument
        """
        Test to see if the SMTP returner sends a message
        """
        ret = {
            "id": "12345",
            "fun": "mytest.func",
            "fun_args": "myfunc args",
            "jid": "54321",
            "return": "The room is on fire as shes fixing her hair",
        }
        options = {
            "username": "",
            "tls": "",
            "from": "",
            "fields": "id,fun,fun_args,jid,return",
            "to": "",
            "host": "",
            "renderer": "jinja|yaml",
            "template": "",
            "password": "",
            "gpgowner": "",
            "subject": "",
        }

        with patch(
            "salt.returners.smtp_return._get_options", MagicMock(return_value=options)
        ), patch.object(SaltCacheLoader, "file_client", MagicMock()):
            smtp.returner(ret)
            self.assertTrue(mocked_smtplib.return_value.sendmail.called)

    if HAS_GNUPG:

        def test_returner(self):
            with patch("salt.returners.smtp_return.gnupg"), patch(
                "salt.returners.smtp_return.smtplib.SMTP"
            ) as mocked_smtplib:
                self._test_returner(mocked_smtplib)

    else:

        def test_returner(self):
            with patch("salt.returners.smtp_return.smtplib.SMTP") as mocked_smtplib:
                self._test_returner(mocked_smtplib)



"""
    tests.pytests.conftest
    ~~~~~~~~~~~~~~~~~~~~~~
"""
import functools
import inspect
import logging
import os
import shutil
import stat

import attr
import pytest
import salt.ext.tornado.ioloop
import salt.utils.files
import salt.utils.platform
from salt.serializers import yaml
from saltfactories.utils import random_string
from saltfactories.utils.ports import get_unused_localhost_port
from tests.support.helpers import get_virtualenv_binary_path
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def salt_minion_id():
    return random_string("minion-")


@pytest.fixture(scope="session")
def salt_sub_minion_id():
    return random_string("sub-minion-")


@pytest.fixture(scope="session")
def sdb_etcd_port():
    return get_unused_localhost_port()


@pytest.fixture(scope="session")
def vault_port():
    return get_unused_localhost_port()


@attr.s(slots=True, frozen=True)
class ReactorEvent:
    sls_path = attr.ib()
    tag = attr.ib()
    event_tag = attr.ib()


@pytest.fixture(scope="session")
def reactor_event(tmp_path_factory):

    reactor_tag = "salt/event/test"
    event_tag = random_string("test/reaction/")
    reactors_dir = tmp_path_factory.mktemp("reactors")
    reactor_test_contents = """
    reactor-test:
      runner.event.send:
        - args:
          - tag: {}
          - data:
              test_reaction: True
    """.format(
        event_tag
    )
    try:
        with pytest.helpers.temp_file(
            "reactor-test.sls", reactor_test_contents, reactors_dir
        ) as reactor_file_path:
            yield ReactorEvent(reactor_file_path, reactor_tag, event_tag)
    finally:
        shutil.rmtree(str(reactors_dir), ignore_errors=True)


@pytest.fixture(scope="session")
def master_id():
    master_id = random_string("master-")
    yield master_id


@pytest.fixture(scope="session")
def salt_master_factory(
    request,
    salt_factories,
    salt_minion_id,
    salt_sub_minion_id,
    base_env_state_tree_root_dir,
    base_env_pillar_tree_root_dir,
    prod_env_state_tree_root_dir,
    prod_env_pillar_tree_root_dir,
    ext_pillar_file_tree_root_dir,
    sdb_etcd_port,
    vault_port,
    reactor_event,
    master_id,
):
    root_dir = salt_factories.get_root_dir_for_daemon(master_id)
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(exist_ok=True)

    with salt.utils.files.fopen(os.path.join(RUNTIME_VARS.CONF_DIR, "master")) as rfh:
        config_defaults = yaml.deserialize(rfh.read())

        tests_known_hosts_file = str(root_dir / "salt_ssh_known_hosts")
        with salt.utils.files.fopen(tests_known_hosts_file, "w") as known_hosts:
            known_hosts.write("")

    config_defaults["root_dir"] = str(root_dir)
    config_defaults["known_hosts_file"] = tests_known_hosts_file
    config_defaults["syndic_master"] = "localhost"
    config_defaults["transport"] = request.config.getoption("--transport")
    config_defaults["reactor"] = [
        {reactor_event.tag: [str(reactor_event.sls_path)]},
    ]

    nodegroups = {
        "min": salt_minion_id,
        "sub_min": salt_sub_minion_id,
        "mins": "N@min or N@sub_min",
        "list_nodegroup": [salt_minion_id, salt_sub_minion_id],
        "multiline_nodegroup": [salt_minion_id, "or", salt_sub_minion_id],
        "one_minion_list": [salt_minion_id],
        "redundant_minions": "N@min or N@mins",
        "nodegroup_loop_a": "N@nodegroup_loop_b",
        "nodegroup_loop_b": "N@nodegroup_loop_a",
        "missing_minion": "L@{},ghostminion".format(salt_minion_id),
        "list_group": "N@multiline_nodegroup",
        "one_list_group": "N@one_minion_list",
        "list_group2": "N@list_nodegroup",
    }
    config_defaults["nodegroups"] = nodegroups
    config_defaults["sdbetcd"] = {
        "driver": "etcd",
        "etcd.host": "127.0.0.1",
        "etcd.port": sdb_etcd_port,
    }
    config_defaults["vault"] = {
        "url": "http://127.0.0.1:{}".format(vault_port),
        "auth": {"method": "token", "token": "testsecret", "uses": 0},
        "policies": ["testpolicy"],
    }

    # Config settings to test `event_return`
    config_defaults["returner_dirs"] = []
    config_defaults["returner_dirs"].append(
        os.path.join(RUNTIME_VARS.FILES, "returners")
    )
    config_defaults["event_return"] = "runtests_noop"
    config_overrides = {"pytest-master": {"log": {"level": "DEBUG"}}}
    ext_pillar = []
    if salt.utils.platform.is_windows():
        ext_pillar.append(
            {"cmd_yaml": "type {}".format(os.path.join(RUNTIME_VARS.FILES, "ext.yaml"))}
        )
    else:
        ext_pillar.append(
            {"cmd_yaml": "cat {}".format(os.path.join(RUNTIME_VARS.FILES, "ext.yaml"))}
        )
    ext_pillar.append(
        {
            "file_tree": {
                "root_dir": str(ext_pillar_file_tree_root_dir),
                "follow_dir_links": False,
                "keep_newline": True,
            }
        }
    )
    config_overrides["pillar_opts"] = True

    # We need to copy the extension modules into the new master root_dir or
    # it will be prefixed by it
    extension_modules_path = str(root_dir / "extension_modules")
    if not os.path.exists(extension_modules_path):
        shutil.copytree(
            os.path.join(RUNTIME_VARS.FILES, "extension_modules"),
            extension_modules_path,
        )

    # Copy the autosign_file to the new  master root_dir
    autosign_file_path = str(root_dir / "autosign_file")
    shutil.copyfile(
        os.path.join(RUNTIME_VARS.FILES, "autosign_file"), autosign_file_path
    )
    # all read, only owner write
    autosign_file_permissions = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
    )
    os.chmod(autosign_file_path, autosign_file_permissions)

    config_overrides.update(
        {
            "ext_pillar": ext_pillar,
            "extension_modules": extension_modules_path,
            "file_roots": {
                "base": [
                    str(base_env_state_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "file", "base"),
                ],
                # Alternate root to test __env__ choices
                "prod": [
                    str(prod_env_state_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "file", "prod"),
                ],
            },
            "pillar_roots": {
                "base": [
                    str(base_env_pillar_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "pillar", "base"),
                ],
                "prod": [str(prod_env_pillar_tree_root_dir)],
            },
        }
    )

    # Let's copy over the test cloud config files and directories into the running master config directory
    for entry in os.listdir(RUNTIME_VARS.CONF_DIR):
        if not entry.startswith("cloud"):
            continue
        source = os.path.join(RUNTIME_VARS.CONF_DIR, entry)
        dest = str(conf_dir / entry)
        if os.path.isdir(source):
            shutil.copytree(source, dest)
        else:
            shutil.copyfile(source, dest)

    factory = salt_factories.salt_master_daemon(
        master_id,
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
    )
    return factory


@pytest.fixture(scope="session")
def salt_minion_factory(salt_master_factory, salt_minion_id, sdb_etcd_port, vault_port):
    with salt.utils.files.fopen(os.path.join(RUNTIME_VARS.CONF_DIR, "minion")) as rfh:
        config_defaults = yaml.deserialize(rfh.read())
    config_defaults["hosts.file"] = os.path.join(RUNTIME_VARS.TMP, "hosts")
    config_defaults["aliases.file"] = os.path.join(RUNTIME_VARS.TMP, "aliases")
    config_defaults["transport"] = salt_master_factory.config["transport"]
    config_defaults["sdbetcd"] = {
        "driver": "etcd",
        "etcd.host": "127.0.0.1",
        "etcd.port": sdb_etcd_port,
    }
    config_defaults["vault"] = {
        "url": "http://127.0.0.1:{}".format(vault_port),
        "auth": {"method": "token", "token": "testsecret", "uses": 0},
        "policies": ["testpolicy"],
    }

    config_overrides = {
        "file_roots": salt_master_factory.config["file_roots"].copy(),
        "pillar_roots": salt_master_factory.config["pillar_roots"].copy(),
    }

    virtualenv_binary = get_virtualenv_binary_path()
    if virtualenv_binary:
        config_overrides["venv_bin"] = virtualenv_binary
    factory = salt_master_factory.salt_minion_daemon(
        salt_minion_id,
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master_factory, factory.id
    )
    return factory


@pytest.fixture(scope="session")
def salt_sub_minion_factory(salt_master_factory, salt_sub_minion_id):
    with salt.utils.files.fopen(
        os.path.join(RUNTIME_VARS.CONF_DIR, "sub_minion")
    ) as rfh:
        config_defaults = yaml.deserialize(rfh.read())
    config_defaults["hosts.file"] = os.path.join(RUNTIME_VARS.TMP, "hosts")
    config_defaults["aliases.file"] = os.path.join(RUNTIME_VARS.TMP, "aliases")
    config_defaults["transport"] = salt_master_factory.config["transport"]

    config_overrides = {
        "file_roots": salt_master_factory.config["file_roots"].copy(),
        "pillar_roots": salt_master_factory.config["pillar_roots"].copy(),
    }

    virtualenv_binary = get_virtualenv_binary_path()
    if virtualenv_binary:
        config_overrides["venv_bin"] = virtualenv_binary
    factory = salt_master_factory.salt_minion_daemon(
        salt_sub_minion_id,
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master_factory, factory.id
    )
    return factory


@pytest.fixture(scope="session")
def salt_proxy_factory(salt_master_factory):
    proxy_minion_id = random_string("proxytest-")

    config_overrides = {
        "file_roots": salt_master_factory.config["file_roots"].copy(),
        "pillar_roots": salt_master_factory.config["pillar_roots"].copy(),
    }

    factory = salt_master_factory.salt_proxy_minion_daemon(
        proxy_minion_id,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
        start_timeout=240,
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master_factory, factory.id
    )
    return factory


@pytest.fixture(scope="session")
def salt_delta_proxy_factory(salt_factories, salt_master_factory):
    proxy_minion_id = random_string("proxytest-")
    root_dir = salt_factories.get_root_dir_for_daemon(proxy_minion_id)
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(parents=True, exist_ok=True)

    config_defaults = {
        "root_dir": str(root_dir),
        "hosts.file": os.path.join(
            RUNTIME_VARS.TMP, "hosts"
        ),  # Do we really need this for these tests?
        "aliases.file": os.path.join(
            RUNTIME_VARS.TMP, "aliases"
        ),  # Do we really need this for these tests?
        "transport": salt_master_factory.config["transport"],
        "user": salt_master_factory.config["user"],
        "metaproxy": "deltaproxy",
        "master": "127.0.0.1",
    }
    factory = salt_master_factory.salt_proxy_minion_daemon(
        proxy_minion_id,
        defaults=config_defaults,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
        start_timeout=240,
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master_factory, factory.id
    )
    return factory


@pytest.fixture
def temp_salt_master(
    request,
    salt_factories,
):
    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    factory = salt_factories.salt_master_daemon(
        random_string("temp-master-"),
        defaults=config_defaults,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
    )
    return factory


@pytest.fixture
def temp_salt_minion(temp_salt_master):
    config_defaults = {
        "open_mode": True,
        "transport": temp_salt_master.config["transport"],
    }
    factory = temp_salt_master.salt_minion_daemon(
        random_string("temp-minion-"),
        defaults=config_defaults,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, temp_salt_master, factory.id
    )
    return factory


@pytest.fixture(scope="session")
def bridge_pytest_and_runtests():
    """
    We're basically overriding the same fixture defined in tests/conftest.py
    """


# ----- Async Test Fixtures ----------------------------------------------------------------------------------------->
# This is based on https://github.com/eukaryote/pytest-tornasync
# The reason why we don't use that pytest plugin instead is because it has
# tornado as a dependency, and we need to use the tornado we ship with salt


def get_test_timeout(pyfuncitem):
    default_timeout = 30
    marker = pyfuncitem.get_closest_marker("timeout")
    if marker:
        return marker.kwargs.get("seconds") or default_timeout
    return default_timeout


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name) and inspect.iscoroutinefunction(obj):
        return list(collector._genfunctions(name, obj))


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    if inspect.iscoroutinefunction(item.obj):
        if "io_loop" not in item.fixturenames:
            # Append the io_loop fixture for the async functions
            item.fixturenames.append("io_loop")


class CoroTestFunction:
    def __init__(self, func, kwargs):
        self.func = func
        self.kwargs = kwargs
        functools.update_wrapper(self, func)

    async def __call__(self):
        ret = await self.func(**self.kwargs)
        return ret


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    if not inspect.iscoroutinefunction(pyfuncitem.obj):
        return

    funcargs = pyfuncitem.funcargs
    testargs = {arg: funcargs[arg] for arg in pyfuncitem._fixtureinfo.argnames}

    try:
        loop = funcargs["io_loop"]
    except KeyError:
        loop = salt.ext.tornado.ioloop.IOLoop.current()

    loop.run_sync(
        CoroTestFunction(pyfuncitem.obj, testargs), timeout=get_test_timeout(pyfuncitem)
    )
    return True


@pytest.fixture
def io_loop():
    """
    Create new io loop for each test, and tear it down after.
    """
    loop = salt.ext.tornado.ioloop.IOLoop()
    loop.make_current()
    try:
        yield loop
    finally:
        loop.clear_current()
        loop.close(all_fds=True)


# <---- Async Test Fixtures ------------------------------------------------------------------------------------------



import salt.modules.mac_pkgutil as mac_pkgutil
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class MacPkgutilTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {mac_pkgutil: {}}

    def test_install(self):
        # Given
        source = "/foo/bar/fubar.pkg"
        package_id = "com.foo.fubar.pkg"

        # When
        with patch("salt.modules.mac_pkgutil.is_installed", return_value=False):
            with patch(
                "salt.modules.mac_pkgutil._install_from_path", return_value=True
            ) as _install_from_path:
                mac_pkgutil.install(source, package_id)

        # Then
        _install_from_path.assert_called_with(source)

    def test_install_already_there(self):
        # Given
        source = "/foo/bar/fubar.pkg"
        package_id = "com.foo.fubar.pkg"

        # When
        with patch("salt.modules.mac_pkgutil.is_installed", return_value=True):
            with patch(
                "salt.modules.mac_pkgutil._install_from_path", return_value=True
            ) as _install_from_path:
                mac_pkgutil.install(source, package_id)

        # Then
        self.assertEqual(_install_from_path.called, 0)



import pytest
import salt.beacons.sensehat as sensehat
from tests.support.mock import MagicMock


@pytest.fixture
def configure_loader_modules():
    return {
        sensehat: {
            "__salt__": {
                "sensehat.get_humidity": MagicMock(return_value=80),
                "sensehat.get_temperature": MagicMock(return_value=30),
                "sensehat.get_pressure": MagicMock(return_value=1500),
            },
        }
    }


def test_non_list_config():
    config = {}

    ret = sensehat.validate(config)

    assert ret == (False, "Configuration for sensehat beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = sensehat.validate(config)

    assert ret == (False, "Configuration for sensehat beacon requires sensors.")


def test_sensehat_humidity_match():

    config = [{"sensors": {"humidity": "70%"}}]

    ret = sensehat.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = sensehat.beacon(config)
    assert ret == [{"tag": "sensehat/humidity", "humidity": 80}]

    # Test without the percent
    config = [{"sensors": {"humidity": "70%"}}]

    ret = sensehat.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = sensehat.beacon(config)
    assert ret == [{"tag": "sensehat/humidity", "humidity": 80}]


def test_sensehat_temperature_match():

    config = [{"sensors": {"temperature": 20}}]

    ret = sensehat.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = sensehat.beacon(config)
    assert ret == [{"tag": "sensehat/temperature", "temperature": 30}]


def test_sensehat_temperature_match_range():

    config = [{"sensors": {"temperature": [20, 29]}}]

    ret = sensehat.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = sensehat.beacon(config)
    assert ret == [{"tag": "sensehat/temperature", "temperature": 30}]


def test_sensehat_pressure_match():

    config = [{"sensors": {"pressure": "1400"}}]

    ret = sensehat.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = sensehat.beacon(config)
    assert ret == [{"tag": "sensehat/pressure", "pressure": 1500}]


def test_sensehat_no_match():

    config = [{"sensors": {"pressure": "1600"}}]

    ret = sensehat.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = sensehat.beacon(config)
    assert ret == []



"""
    :codeauthor: Bo Maryniuk <bo@suse.de>
"""
import imp
import os

from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    from zypp_plugin import BogusIO

    HAS_ZYPP_PLUGIN = True
except ImportError:
    HAS_ZYPP_PLUGIN = False

BUILTINS_OPEN = "builtins.open"

ZYPPNOTIFY_FILE = os.path.sep.join(
    os.path.dirname(__file__).split(os.path.sep)[:-2]
    + ["scripts", "suse", "zypper", "plugins", "commit", "zyppnotify"]
)


@skipIf(not HAS_ZYPP_PLUGIN, "zypp_plugin is missing.")
class ZyppPluginsTestCase(TestCase):
    """
    Test shipped libzypp plugins.
    """

    @skipIf(
        not os.path.exists(ZYPPNOTIFY_FILE),
        "Required file '{}' does not exist.".format(ZYPPNOTIFY_FILE),
    )
    def test_drift_detector(self):
        """
        Test drift detector for a correct cookie file.
        Returns:

        """
        zyppnotify = imp.load_source("zyppnotify", ZYPPNOTIFY_FILE)
        drift = zyppnotify.DriftDetector()
        drift._get_mtime = MagicMock(return_value=123)
        drift._get_checksum = MagicMock(return_value="deadbeef")
        bogus_io = BogusIO()
        with patch(BUILTINS_OPEN, bogus_io):
            drift.PLUGINEND(None, None)
        self.assertEqual(str(bogus_io), "deadbeef 123\n")
        self.assertEqual(bogus_io.mode, "w")
        self.assertEqual(bogus_io.path, "/var/cache/salt/minion/rpmdb.cookie")



"""
Beacon to fire event when we notice a AIX user is locked due to many failed login attempts.

.. versionadded:: 2018.3.0

:depends: none
"""
import logging

log = logging.getLogger(__name__)

__virtualname__ = "aix_account"


def __virtual__():
    """
    Only load if kernel is AIX
    """
    if __grains__["kernel"] == "AIX":
        return __virtualname__

    return (
        False,
        "The aix_account beacon module failed to load: only available on AIX systems.",
    )


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for aix_account beacon should be a dictionary
    if not isinstance(config, dict):
        return False, "Configuration for aix_account beacon must be a dict."
    if "user" not in config:
        return (
            False,
            "Configuration for aix_account beacon must include a user or ALL for all users.",
        )
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Checks for locked accounts due to too many invalid login attempts, 3 or higher.

    .. code-block:: yaml

        beacons:
          aix_account:
            user: ALL
            interval: 120

    """

    ret = []

    user = config["user"]

    locked_accounts = __salt__["shadow.login_failures"](user)
    ret.append({"accounts": locked_accounts})

    return ret



import io
import stat
from functools import wraps

import pytest
import salt.config
import salt.daemons.masterapi as masterapi
import salt.utils.platform
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


def gen_permissions(owner="", group="", others=""):
    """
    Helper method to generate file permission bits
    Usage: gen_permissions('rw', 'r', 'r')
    """
    ret = 0
    for c in owner:
        ret |= getattr(stat, "S_I{}USR".format(c.upper()), 0)
    for c in group:
        ret |= getattr(stat, "S_I{}GRP".format(c.upper()), 0)
    for c in others:
        ret |= getattr(stat, "S_I{}OTH".format(c.upper()), 0)
    return ret


def patch_check_permissions(uid=1, groups=None, is_windows=False, permissive_pki=False):
    if not groups:
        groups = [uid]

    def decorator(func):
        @wraps(func)
        def wrapper(self):
            self.auto_key.opts["permissive_pki_access"] = permissive_pki
            if salt.utils.platform.is_windows():
                with patch(
                    "salt.utils.platform.is_windows", MagicMock(return_value=True)
                ):
                    func(self)
            else:
                with patch("os.stat", self.os_stat_mock), patch(
                    "os.getuid", MagicMock(return_value=uid)
                ), patch(
                    "salt.utils.user.get_gid_list", MagicMock(return_value=groups)
                ), patch(
                    "salt.utils.platform.is_windows", MagicMock(return_value=is_windows)
                ):
                    func(self)

        return wrapper

    return decorator


class AutoKeyTest(TestCase):
    """
    Tests for the salt.daemons.masterapi.AutoKey class
    """

    def setUp(self):
        opts = salt.config.master_config(None)
        opts["user"] = "test_user"
        self.auto_key = masterapi.AutoKey(opts)
        self.stats = {}

    def os_stat_mock(self, filename):
        fmode = MagicMock()
        fstats = self.stats.get(filename, {})
        fmode.st_mode = fstats.get("mode", 0)
        fmode.st_gid = fstats.get("gid", 0)
        return fmode

    @patch_check_permissions(uid=0, is_windows=True)
    def test_check_permissions_windows(self):
        """
        Assert that all files are accepted on windows
        """
        self.stats["testfile"] = {
            "mode": gen_permissions("rwx", "rwx", "rwx"),
            "gid": 2,
        }
        self.assertTrue(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions(permissive_pki=True)
    def test_check_permissions_others_can_write(self):
        """
        Assert that no file is accepted, when others can write to it
        """
        self.stats["testfile"] = {"mode": gen_permissions("", "", "w"), "gid": 1}
        if salt.utils.platform.is_windows():
            self.assertTrue(self.auto_key.check_permissions("testfile"))
        else:
            self.assertFalse(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions()
    def test_check_permissions_group_can_write_not_permissive(self):
        """
        Assert that a file is accepted, when group can write to it and
        permissive_pki_access=False
        """
        self.stats["testfile"] = {"mode": gen_permissions("w", "w", ""), "gid": 1}
        if salt.utils.platform.is_windows():
            self.assertTrue(self.auto_key.check_permissions("testfile"))
        else:
            self.assertFalse(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions(permissive_pki=True)
    def test_check_permissions_group_can_write_permissive(self):
        """
        Assert that a file is accepted, when group can write to it and
        permissive_pki_access=True
        """
        self.stats["testfile"] = {"mode": gen_permissions("w", "w", ""), "gid": 1}
        self.assertTrue(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions(uid=0, permissive_pki=True)
    def test_check_permissions_group_can_write_permissive_root_in_group(self):
        """
        Assert that a file is accepted, when group can write to it,
        permissive_pki_access=False, salt is root and in the file owning group
        """
        self.stats["testfile"] = {"mode": gen_permissions("w", "w", ""), "gid": 0}
        self.assertTrue(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions(uid=0, permissive_pki=True)
    def test_check_permissions_group_can_write_permissive_root_not_in_group(self):
        """
        Assert that no file is accepted, when group can write to it,
        permissive_pki_access=False, salt is root and **not** in the file owning
        group
        """
        self.stats["testfile"] = {"mode": gen_permissions("w", "w", ""), "gid": 1}
        if salt.utils.platform.is_windows():
            self.assertTrue(self.auto_key.check_permissions("testfile"))
        else:
            self.assertFalse(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions()
    def test_check_permissions_only_owner_can_write(self):
        """
        Assert that a file is accepted, when only the owner can write to it
        """
        self.stats["testfile"] = {"mode": gen_permissions("w", "", ""), "gid": 1}
        self.assertTrue(self.auto_key.check_permissions("testfile"))

    @patch_check_permissions(uid=0)
    def test_check_permissions_only_owner_can_write_root(self):
        """
        Assert that a file is accepted, when only the owner can write to it and salt is root
        """
        self.stats["testfile"] = {"mode": gen_permissions("w", "", ""), "gid": 0}
        self.assertTrue(self.auto_key.check_permissions("testfile"))

    def _test_check_autosign_grains(
        self,
        test_func,
        file_content="test_value",
        file_name="test_grain",
        autosign_grains_dir="test_dir",
        permissions_ret=True,
    ):
        """
        Helper function for testing autosign_grains().

        Patches ``os.walk`` to return only ``file_name`` and ``salt.utils.files.fopen`` to open a
        mock file with ``file_content`` as content. Optionally sets ``opts`` values.
        Then executes test_func. The ``os.walk`` and ``salt.utils.files.fopen`` mock objects
        are passed to the function as arguments.
        """
        if autosign_grains_dir:
            self.auto_key.opts["autosign_grains_dir"] = autosign_grains_dir
        mock_file = io.StringIO(file_content)
        mock_dirs = [(None, None, [file_name])]

        with patch("os.walk", MagicMock(return_value=mock_dirs)) as mock_walk, patch(
            "salt.utils.files.fopen", MagicMock(return_value=mock_file)
        ) as mock_open, patch(
            "salt.daemons.masterapi.AutoKey.check_permissions",
            MagicMock(return_value=permissions_ret),
        ) as mock_permissions:
            test_func(mock_walk, mock_open, mock_permissions)

    def test_check_autosign_grains_no_grains(self):
        """
        Asserts that autosigning from grains fails when no grain values are passed.
        """

        def test_func(mock_walk, mock_open, mock_permissions):
            self.assertFalse(self.auto_key.check_autosign_grains(None))
            self.assertEqual(mock_walk.call_count, 0)
            self.assertEqual(mock_open.call_count, 0)
            self.assertEqual(mock_permissions.call_count, 0)

            self.assertFalse(self.auto_key.check_autosign_grains({}))
            self.assertEqual(mock_walk.call_count, 0)
            self.assertEqual(mock_open.call_count, 0)
            self.assertEqual(mock_permissions.call_count, 0)

        self._test_check_autosign_grains(test_func)

    def test_check_autosign_grains_no_autosign_grains_dir(self):
        """
        Asserts that autosigning from grains fails when the \'autosign_grains_dir\' config option
        is undefined.
        """

        def test_func(mock_walk, mock_open, mock_permissions):
            self.assertFalse(
                self.auto_key.check_autosign_grains({"test_grain": "test_value"})
            )
            self.assertEqual(mock_walk.call_count, 0)
            self.assertEqual(mock_open.call_count, 0)
            self.assertEqual(mock_permissions.call_count, 0)

        self._test_check_autosign_grains(test_func, autosign_grains_dir=None)

    def test_check_autosign_grains_accept(self):
        """
        Asserts that autosigning from grains passes when a matching grain value is in an
        autosign_grain file.
        """

        def test_func(*args):
            self.assertTrue(
                self.auto_key.check_autosign_grains({"test_grain": "test_value"})
            )

        file_content = "#test_ignore\ntest_value"
        self._test_check_autosign_grains(test_func, file_content=file_content)

    def test_check_autosign_grains_accept_not(self):
        """
        Asserts that autosigning from grains fails when the grain value is not in the
        autosign_grain files.
        """

        def test_func(*args):
            self.assertFalse(
                self.auto_key.check_autosign_grains({"test_grain": "test_invalid"})
            )

        file_content = "#test_invalid\ntest_value"
        self._test_check_autosign_grains(test_func, file_content=file_content)

    def test_check_autosign_grains_invalid_file_permissions(self):
        """
        Asserts that autosigning from grains fails when the grain file has the wrong permissions.
        """

        def test_func(*args):
            self.assertFalse(
                self.auto_key.check_autosign_grains({"test_grain": "test_value"})
            )

        file_content = "#test_ignore\ntest_value"
        self._test_check_autosign_grains(
            test_func, file_content=file_content, permissions_ret=False
        )


class LocalFuncsTestCase(TestCase):
    """
    TestCase for salt.daemons.masterapi.LocalFuncs class
    """

    def setUp(self):
        opts = salt.config.master_config(None)
        self.local_funcs = masterapi.LocalFuncs(opts, "test-key")
        self.addCleanup(self.local_funcs.destroy)
        self.addCleanup(delattr, self, "local_funcs")

    # runner tests

    @pytest.mark.slow_test
    def test_runner_token_not_authenticated(self):
        """
        Asserts that a TokenAuthenticationError is returned when the token can't authenticate.
        """
        mock_ret = {
            "error": {
                "name": "TokenAuthenticationError",
                "message": 'Authentication failure of type "token" occurred.',
            }
        }
        ret = self.local_funcs.runner({"token": "asdfasdfasdfasdf"})
        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_runner_token_authorization_error(self):
        """
        Asserts that a TokenAuthenticationError is returned when the token authenticates, but is
        not authorized.
        """
        token = "asdfasdfasdfasdf"
        load = {"token": token, "fun": "test.arg", "kwarg": {}}
        mock_token = {"token": token, "eauth": "foo", "name": "test"}
        mock_ret = {
            "error": {
                "name": "TokenAuthenticationError",
                "message": (
                    'Authentication failure of type "token" occurred for user test.'
                ),
            }
        }

        with patch(
            "salt.auth.LoadAuth.authenticate_token", MagicMock(return_value=mock_token)
        ), patch("salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=[])):
            ret = self.local_funcs.runner(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_runner_token_salt_invocation_error(self):
        """
        Asserts that a SaltInvocationError is returned when the token authenticates, but the
        command is malformed.
        """
        token = "asdfasdfasdfasdf"
        load = {"token": token, "fun": "badtestarg", "kwarg": {}}
        mock_token = {"token": token, "eauth": "foo", "name": "test"}
        mock_ret = {
            "error": {
                "name": "SaltInvocationError",
                "message": "A command invocation error occurred: Check syntax.",
            }
        }

        with patch(
            "salt.auth.LoadAuth.authenticate_token", MagicMock(return_value=mock_token)
        ), patch(
            "salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=["testing"])
        ):
            ret = self.local_funcs.runner(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_runner_eauth_not_authenticated(self):
        """
        Asserts that an EauthAuthenticationError is returned when the user can't authenticate.
        """
        mock_ret = {
            "error": {
                "name": "EauthAuthenticationError",
                "message": (
                    'Authentication failure of type "eauth" occurred for user UNKNOWN.'
                ),
            }
        }
        ret = self.local_funcs.runner({"eauth": "foo"})
        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_runner_eauth_authorization_error(self):
        """
        Asserts that an EauthAuthenticationError is returned when the user authenticates, but is
        not authorized.
        """
        load = {"eauth": "foo", "username": "test", "fun": "test.arg", "kwarg": {}}
        mock_ret = {
            "error": {
                "name": "EauthAuthenticationError",
                "message": (
                    'Authentication failure of type "eauth" occurred for user test.'
                ),
            }
        }
        with patch(
            "salt.auth.LoadAuth.authenticate_eauth", MagicMock(return_value=True)
        ), patch("salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=[])):
            ret = self.local_funcs.runner(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_runner_eauth_salt_invocation_error(self):
        """
        Asserts that an EauthAuthenticationError is returned when the user authenticates, but the
        command is malformed.
        """
        load = {
            "eauth": "foo",
            "username": "test",
            "fun": "bad.test.arg.func",
            "kwarg": {},
        }
        mock_ret = {
            "error": {
                "name": "SaltInvocationError",
                "message": "A command invocation error occurred: Check syntax.",
            }
        }
        with patch(
            "salt.auth.LoadAuth.authenticate_eauth", MagicMock(return_value=True)
        ), patch(
            "salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=["testing"])
        ):
            ret = self.local_funcs.runner(load)

        self.assertDictEqual(mock_ret, ret)

    # wheel tests

    @pytest.mark.slow_test
    def test_wheel_token_not_authenticated(self):
        """
        Asserts that a TokenAuthenticationError is returned when the token can't authenticate.
        """
        mock_ret = {
            "error": {
                "name": "TokenAuthenticationError",
                "message": 'Authentication failure of type "token" occurred.',
            }
        }
        ret = self.local_funcs.wheel({"token": "asdfasdfasdfasdf"})
        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_wheel_token_authorization_error(self):
        """
        Asserts that a TokenAuthenticationError is returned when the token authenticates, but is
        not authorized.
        """
        token = "asdfasdfasdfasdf"
        load = {"token": token, "fun": "test.arg", "kwarg": {}}
        mock_token = {"token": token, "eauth": "foo", "name": "test"}
        mock_ret = {
            "error": {
                "name": "TokenAuthenticationError",
                "message": (
                    'Authentication failure of type "token" occurred for user test.'
                ),
            }
        }

        with patch(
            "salt.auth.LoadAuth.authenticate_token", MagicMock(return_value=mock_token)
        ), patch("salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=[])):
            ret = self.local_funcs.wheel(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_wheel_token_salt_invocation_error(self):
        """
        Asserts that a SaltInvocationError is returned when the token authenticates, but the
        command is malformed.
        """
        token = "asdfasdfasdfasdf"
        load = {"token": token, "fun": "badtestarg", "kwarg": {}}
        mock_token = {"token": token, "eauth": "foo", "name": "test"}
        mock_ret = {
            "error": {
                "name": "SaltInvocationError",
                "message": "A command invocation error occurred: Check syntax.",
            }
        }

        with patch(
            "salt.auth.LoadAuth.authenticate_token", MagicMock(return_value=mock_token)
        ), patch(
            "salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=["testing"])
        ):
            ret = self.local_funcs.wheel(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_wheel_eauth_not_authenticated(self):
        """
        Asserts that an EauthAuthenticationError is returned when the user can't authenticate.
        """
        mock_ret = {
            "error": {
                "name": "EauthAuthenticationError",
                "message": (
                    'Authentication failure of type "eauth" occurred for user UNKNOWN.'
                ),
            }
        }
        ret = self.local_funcs.wheel({"eauth": "foo"})
        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_wheel_eauth_authorization_error(self):
        """
        Asserts that an EauthAuthenticationError is returned when the user authenticates, but is
        not authorized.
        """
        load = {"eauth": "foo", "username": "test", "fun": "test.arg", "kwarg": {}}
        mock_ret = {
            "error": {
                "name": "EauthAuthenticationError",
                "message": (
                    'Authentication failure of type "eauth" occurred for user test.'
                ),
            }
        }
        with patch(
            "salt.auth.LoadAuth.authenticate_eauth", MagicMock(return_value=True)
        ), patch("salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=[])):
            ret = self.local_funcs.wheel(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_wheel_eauth_salt_invocation_error(self):
        """
        Asserts that an EauthAuthenticationError is returned when the user authenticates, but the
        command is malformed.
        """
        load = {
            "eauth": "foo",
            "username": "test",
            "fun": "bad.test.arg.func",
            "kwarg": {},
        }
        mock_ret = {
            "error": {
                "name": "SaltInvocationError",
                "message": "A command invocation error occurred: Check syntax.",
            }
        }
        with patch(
            "salt.auth.LoadAuth.authenticate_eauth", MagicMock(return_value=True)
        ), patch(
            "salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=["testing"])
        ):
            ret = self.local_funcs.wheel(load)

        self.assertDictEqual(mock_ret, ret)

    @pytest.mark.slow_test
    def test_wheel_user_not_authenticated(self):
        """
        Asserts that an UserAuthenticationError is returned when the user can't authenticate.
        """
        mock_ret = {
            "error": {
                "name": "UserAuthenticationError",
                "message": (
                    'Authentication failure of type "user" occurred for user UNKNOWN.'
                ),
            }
        }
        ret = self.local_funcs.wheel({})
        self.assertDictEqual(mock_ret, ret)

    # publish tests

    @pytest.mark.slow_test
    def test_publish_user_is_blacklisted(self):
        """
        Asserts that an AuthorizationError is returned when the user has been blacklisted.
        """
        mock_ret = {
            "error": {
                "name": "AuthorizationError",
                "message": "Authorization error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=True)
        ):
            self.assertEqual(
                mock_ret, self.local_funcs.publish({"user": "foo", "fun": "test.arg"})
            )

    @pytest.mark.slow_test
    def test_publish_cmd_blacklisted(self):
        """
        Asserts that an AuthorizationError is returned when the command has been blacklisted.
        """
        mock_ret = {
            "error": {
                "name": "AuthorizationError",
                "message": "Authorization error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=True)
        ):
            self.assertEqual(
                mock_ret, self.local_funcs.publish({"user": "foo", "fun": "test.arg"})
            )

    @pytest.mark.slow_test
    def test_publish_token_not_authenticated(self):
        """
        Asserts that an AuthenticationError is returned when the token can't authenticate.
        """
        load = {
            "user": "foo",
            "fun": "test.arg",
            "tgt": "test_minion",
            "kwargs": {"token": "asdfasdfasdfasdf"},
        }
        mock_ret = {
            "error": {
                "name": "AuthenticationError",
                "message": "Authentication error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))

    @pytest.mark.slow_test
    def test_publish_token_authorization_error(self):
        """
        Asserts that an AuthorizationError is returned when the token authenticates, but is not
        authorized.
        """
        token = "asdfasdfasdfasdf"
        load = {
            "user": "foo",
            "fun": "test.arg",
            "tgt": "test_minion",
            "arg": "bar",
            "kwargs": {"token": token},
        }
        mock_token = {"token": token, "eauth": "foo", "name": "test"}
        mock_ret = {
            "error": {
                "name": "AuthorizationError",
                "message": "Authorization error occurred.",
            }
        }

        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.auth.LoadAuth.authenticate_token", MagicMock(return_value=mock_token)
        ), patch(
            "salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=[])
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))

    @pytest.mark.slow_test
    def test_publish_eauth_not_authenticated(self):
        """
        Asserts that an AuthenticationError is returned when the user can't authenticate.
        """
        load = {
            "user": "test",
            "fun": "test.arg",
            "tgt": "test_minion",
            "kwargs": {"eauth": "foo"},
        }
        mock_ret = {
            "error": {
                "name": "AuthenticationError",
                "message": "Authentication error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))

    @pytest.mark.slow_test
    def test_publish_eauth_authorization_error(self):
        """
        Asserts that an AuthorizationError is returned when the user authenticates, but is not
        authorized.
        """
        load = {
            "user": "test",
            "fun": "test.arg",
            "tgt": "test_minion",
            "kwargs": {"eauth": "foo"},
            "arg": "bar",
        }
        mock_ret = {
            "error": {
                "name": "AuthorizationError",
                "message": "Authorization error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.auth.LoadAuth.authenticate_eauth", MagicMock(return_value=True)
        ), patch(
            "salt.auth.LoadAuth.get_auth_list", MagicMock(return_value=[])
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))

    @pytest.mark.slow_test
    def test_publish_user_not_authenticated(self):
        """
        Asserts that an AuthenticationError is returned when the user can't authenticate.
        """
        load = {"user": "test", "fun": "test.arg", "tgt": "test_minion"}
        mock_ret = {
            "error": {
                "name": "AuthenticationError",
                "message": "Authentication error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))

    @pytest.mark.slow_test
    def test_publish_user_authenticated_missing_auth_list(self):
        """
        Asserts that an AuthenticationError is returned when the user has an effective user id and is
        authenticated, but the auth_list is empty.
        """
        load = {
            "user": "test",
            "fun": "test.arg",
            "tgt": "test_minion",
            "kwargs": {"user": "test"},
            "arg": "foo",
        }
        mock_ret = {
            "error": {
                "name": "AuthenticationError",
                "message": "Authentication error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.auth.LoadAuth.authenticate_key",
            MagicMock(return_value="fake-user-key"),
        ), patch(
            "salt.utils.master.get_values_of_matching_keys", MagicMock(return_value=[])
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))

    @pytest.mark.slow_test
    def test_publish_user_authorization_error(self):
        """
        Asserts that an AuthorizationError is returned when the user authenticates, but is not
        authorized.
        """
        load = {
            "user": "test",
            "fun": "test.arg",
            "tgt": "test_minion",
            "kwargs": {"user": "test"},
            "arg": "foo",
        }
        mock_ret = {
            "error": {
                "name": "AuthorizationError",
                "message": "Authorization error occurred.",
            }
        }
        with patch(
            "salt.acl.PublisherACL.user_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.acl.PublisherACL.cmd_is_blacklisted", MagicMock(return_value=False)
        ), patch(
            "salt.auth.LoadAuth.authenticate_key",
            MagicMock(return_value="fake-user-key"),
        ), patch(
            "salt.utils.master.get_values_of_matching_keys",
            MagicMock(return_value=["test"]),
        ), patch(
            "salt.utils.minions.CkMinions.auth_check", MagicMock(return_value=False)
        ):
            self.assertEqual(mock_ret, self.local_funcs.publish(load))


class FakeCache:
    def __init__(self):
        self.data = {}

    def store(self, bank, key, value):
        self.data[bank, key] = value

    def fetch(self, bank, key):
        return self.data[bank, key]


class RemoteFuncsTestCase(TestCase):
    """
    TestCase for salt.daemons.masterapi.RemoteFuncs class
    """

    def setUp(self):
        opts = AdaptedConfigurationTestCaseMixin.get_temp_config("master")
        salt.cache.MemCache.data.clear()
        self.funcs = masterapi.RemoteFuncs(opts)
        self.funcs.cache = FakeCache()
        self.addCleanup(self.funcs.destroy)
        self.addCleanup(delattr, self, "funcs")

    @pytest.mark.slow_test
    def test_mine_get(self, tgt_type_key="tgt_type"):
        """
        Asserts that ``mine_get`` gives the expected results.

        Actually this only tests that:

        - the correct check minions method is called
        - the correct cache key is subsequently used
        """
        self.funcs.cache.store(
            "minions/webserver", "mine", dict(ip_addr="2001:db8::1:3")
        )
        with patch(
            "salt.utils.minions.CkMinions._check_compound_minions",
            MagicMock(return_value=(dict(minions=["webserver"], missing=[]))),
        ):
            ret = self.funcs._mine_get(
                {
                    "id": "requester_minion",
                    "tgt": "G@roles:web",
                    "fun": "ip_addr",
                    tgt_type_key: "compound",
                }
            )
        self.assertDictEqual(ret, dict(webserver="2001:db8::1:3"))

    @pytest.mark.slow_test
    def test_mine_get_pre_nitrogen_compat(self):
        """
        Asserts that pre-Nitrogen API key ``expr_form`` is still accepted.

        This is what minions before Nitrogen would issue.
        """
        self.test_mine_get(tgt_type_key="expr_form")

    @pytest.mark.slow_test
    def test_mine_get_dict_str(self, tgt_type_key="tgt_type"):
        """
        Asserts that ``mine_get`` gives the expected results when request
        is a comma-separated list.

        Actually this only tests that:

        - the correct check minions method is called
        - the correct cache key is subsequently used
        """
        self.funcs.cache.store(
            "minions/webserver",
            "mine",
            dict(ip_addr="2001:db8::1:3", ip4_addr="127.0.0.1"),
        )
        with patch(
            "salt.utils.minions.CkMinions._check_compound_minions",
            MagicMock(return_value=(dict(minions=["webserver"], missing=[]))),
        ):
            ret = self.funcs._mine_get(
                {
                    "id": "requester_minion",
                    "tgt": "G@roles:web",
                    "fun": "ip_addr,ip4_addr",
                    tgt_type_key: "compound",
                }
            )
        self.assertDictEqual(
            ret,
            dict(
                ip_addr=dict(webserver="2001:db8::1:3"),
                ip4_addr=dict(webserver="127.0.0.1"),
            ),
        )

    @pytest.mark.slow_test
    def test_mine_get_dict_list(self, tgt_type_key="tgt_type"):
        """
        Asserts that ``mine_get`` gives the expected results when request
        is a list.

        Actually this only tests that:

        - the correct check minions method is called
        - the correct cache key is subsequently used
        """
        self.funcs.cache.store(
            "minions/webserver",
            "mine",
            dict(ip_addr="2001:db8::1:3", ip4_addr="127.0.0.1"),
        )
        with patch(
            "salt.utils.minions.CkMinions._check_compound_minions",
            MagicMock(return_value=(dict(minions=["webserver"], missing=[]))),
        ):
            ret = self.funcs._mine_get(
                {
                    "id": "requester_minion",
                    "tgt": "G@roles:web",
                    "fun": ["ip_addr", "ip4_addr"],
                    tgt_type_key: "compound",
                }
            )
        self.assertDictEqual(
            ret,
            dict(
                ip_addr=dict(webserver="2001:db8::1:3"),
                ip4_addr=dict(webserver="127.0.0.1"),
            ),
        )

    @pytest.mark.slow_test
    def test_mine_get_acl_allowed(self):
        """
        Asserts that ``mine_get`` gives the expected results when this is allowed
        in the client-side ACL that was stored in the mine data.
        """
        self.funcs.cache.store(
            "minions/webserver",
            "mine",
            {
                "ip_addr": {
                    salt.utils.mine.MINE_ITEM_ACL_DATA: "2001:db8::1:4",
                    salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
                    "allow_tgt": "requester_minion",
                    "allow_tgt_type": "glob",
                },
            },
        )
        # The glob check is for the resolution of the allow_tgt
        # The compound check is for the resolution of the tgt in the mine_get request.
        with patch(
            "salt.utils.minions.CkMinions._check_glob_minions",
            MagicMock(return_value={"minions": ["requester_minion"], "missing": []}),
        ), patch(
            "salt.utils.minions.CkMinions._check_compound_minions",
            MagicMock(return_value={"minions": ["webserver"], "missing": []}),
        ):
            ret = self.funcs._mine_get(
                {
                    "id": "requester_minion",
                    "tgt": "anything",
                    "tgt_type": "compound",
                    "fun": ["ip_addr"],
                }
            )
        self.assertDictEqual(ret, {"ip_addr": {"webserver": "2001:db8::1:4"}})

    @pytest.mark.slow_test
    def test_mine_get_acl_rejected(self):
        """
        Asserts that ``mine_get`` gives the expected results when this is rejected
        in the client-side ACL that was stored in the mine data. This results in
        no data being sent back (just as if the entry wouldn't exist).
        """
        self.funcs.cache.store(
            "minions/webserver",
            "mine",
            {
                "ip_addr": {
                    salt.utils.mine.MINE_ITEM_ACL_DATA: "2001:db8::1:4",
                    salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
                    "allow_tgt": "not_requester_minion",
                    "allow_tgt_type": "glob",
                }
            },
        )
        # The glob check is for the resolution of the allow_tgt
        # The compound check is for the resolution of the tgt in the mine_get request.
        with patch(
            "salt.utils.minions.CkMinions._check_glob_minions",
            MagicMock(
                return_value={"minions": ["not_requester_minion"], "missing": []}
            ),
        ), patch(
            "salt.utils.minions.CkMinions._check_compound_minions",
            MagicMock(return_value={"minions": ["webserver"], "missing": []}),
        ):
            ret = self.funcs._mine_get(
                {
                    "id": "requester_minion",
                    "tgt": "anything",
                    "tgt_type": "compound",
                    "fun": ["ip_addr"],
                }
            )
        self.assertDictEqual(ret, {})



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import salt.modules.firewalld as firewalld
from tests.support.helpers import dedent
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class FirewalldTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.firewalld
    """

    def setup_loader_modules(self):
        return {firewalld: {}}

    def test_version(self):
        """
        Test for Return version from firewall-cmd
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=2):
            self.assertEqual(firewalld.version(), 2)

    def test_default_zone(self):
        """
        Test for Print default zone for connections and interfaces
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            self.assertEqual(firewalld.default_zone(), "A")

    def test_list_zones(self):
        """
        Test for List everything added for or enabled in all zones
        """
        # pylint: disable=trailing-whitespace
        firewall_cmd_ret = dedent(
            """\
                nm-shared
                  target: ACCEPT
                  icmp-block-inversion: no
                  interfaces: 
                  sources: 
                  services: dhcp dns ssh
                  ports: 
                  protocols: icmp ipv6-icmp
                  masquerade: no
                  forward-ports: 
                  source-ports: 
                  icmp-blocks: 
                  rich rules: 
                \trule priority="32767" reject

                public
                  target: default
                  icmp-block-inversion: no
                  interfaces: 
                  sources: 
                  services: cockpit dhcpv6-client ssh
                  ports: 
                  protocols: 
                  masquerade: no
                  forward-ports: 
                  source-ports: 
                  icmp-blocks: 
                  rich rules:
                """
        )
        # pylint: enable=trailing-whitespace
        ret = {
            "nm-shared": {
                "forward-ports": [""],
                "icmp-block-inversion": ["no"],
                "icmp-blocks": [""],
                "interfaces": [""],
                "masquerade": ["no"],
                "ports": [""],
                "protocols": ["icmp ipv6-icmp"],
                "rich rules": ["", 'rule priority="32767" reject'],
                "services": ["dhcp dns ssh"],
                "source-ports": [""],
                "sources": [""],
                "target": ["ACCEPT"],
            },
            "public": {
                "forward-ports": [""],
                "icmp-block-inversion": ["no"],
                "icmp-blocks": [""],
                "interfaces": [""],
                "masquerade": ["no"],
                "ports": [""],
                "protocols": [""],
                "rich rules": [""],
                "services": ["cockpit dhcpv6-client ssh"],
                "source-ports": [""],
                "sources": [""],
                "target": ["default"],
            },
        }

        with patch.object(firewalld, "__firewall_cmd", return_value=firewall_cmd_ret):
            self.assertEqual(firewalld.list_zones(), ret)

    def test_list_zones_empty_response(self):
        """
        Test list_zones if firewall-cmd call returns nothing
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=""):
            self.assertEqual(firewalld.list_zones(), {})

    def test_get_zones(self):
        """
        Test for Print predefined zones
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            self.assertEqual(firewalld.get_zones(), ["A"])

    def test_get_services(self):
        """
        Test for Print predefined services
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            self.assertEqual(firewalld.get_services(), ["A"])

    def test_get_icmp_types(self):
        """
        Test for Print predefined icmptypes
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            self.assertEqual(firewalld.get_icmp_types(), ["A"])

    def test_new_zone(self):
        """
        Test for Add a new zone
        """
        with patch.object(firewalld, "__mgmt", return_value="success"):
            mock = MagicMock(return_value="A")
            with patch.object(firewalld, "__firewall_cmd", mock):
                self.assertEqual(firewalld.new_zone("zone"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.new_zone("zone"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.new_zone("zone", False), "A")

    def test_delete_zone(self):
        """
        Test for Delete an existing zone
        """
        with patch.object(firewalld, "__mgmt", return_value="success"):
            with patch.object(firewalld, "__firewall_cmd", return_value="A"):
                self.assertEqual(firewalld.delete_zone("zone"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.delete_zone("zone"), "A")

        mock = MagicMock(return_value="A")
        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.delete_zone("zone", False), "A")

    def test_set_default_zone(self):
        """
        Test for Set default zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            self.assertEqual(firewalld.set_default_zone("zone"), "A")

    def test_new_service(self):
        """
        Test for Add a new service
        """
        with patch.object(firewalld, "__mgmt", return_value="success"):
            mock = MagicMock(return_value="A")
            with patch.object(firewalld, "__firewall_cmd", return_value="A"):
                self.assertEqual(firewalld.new_service("zone"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.new_service("zone"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.new_service("zone", False), "A")

    def test_delete_service(self):
        """
        Test for Delete an existing service
        """
        with patch.object(firewalld, "__mgmt", return_value="success"):
            mock = MagicMock(return_value="A")
            with patch.object(firewalld, "__firewall_cmd", return_value="A"):
                self.assertEqual(firewalld.delete_service("name"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.delete_service("name"), "A")

        with patch.object(firewalld, "__mgmt", return_value="A"):
            self.assertEqual(firewalld.delete_service("name", False), "A")

    def test_list_all(self):
        """
        Test for List everything added for or enabled in a zone
        """
        # pylint: disable=trailing-whitespace
        firewall_cmd_ret = dedent(
            """\
            public
              target: default
              icmp-block-inversion: no
              interfaces: eth0
              sources: 
              services: cockpit dhcpv6-client ssh
              ports: 
              protocols: 
              masquerade: no
              forward-ports: 
              source-ports: 
              icmp-blocks: 
              rich rules: 
            """
        )
        # pylint: enable=trailing-whitespace
        ret = {
            "public": {
                "forward-ports": [""],
                "icmp-block-inversion": ["no"],
                "icmp-blocks": [""],
                "interfaces": ["eth0"],
                "masquerade": ["no"],
                "ports": [""],
                "protocols": [""],
                "rich rules": [""],
                "services": ["cockpit dhcpv6-client ssh"],
                "source-ports": [""],
                "sources": [""],
                "target": ["default"],
            }
        }
        with patch.object(firewalld, "__firewall_cmd", return_value=firewall_cmd_ret):
            self.assertEqual(firewalld.list_all(), ret)

    def test_list_all_empty_response(self):
        """
        Test list_all if firewall-cmd call returns nothing
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=""):
            self.assertEqual(firewalld.list_all(), {})

    def test_list_services(self):
        """
        Test for List services added for zone as a space separated list.
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=""):
            self.assertEqual(firewalld.list_services(), [])

    def test_add_service(self):
        """
        Test for Add a service for zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=""):
            self.assertEqual(firewalld.add_service("name"), "")

    def test_remove_service(self):
        """
        Test for Remove a service from zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=""):
            self.assertEqual(firewalld.remove_service("name"), "")

    def test_add_masquerade(self):
        """
        Test for adding masquerade
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(firewalld.add_masquerade("name"), "success")

    def test_remove_masquerade(self):
        """
        Test for removing masquerade
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(firewalld.remove_masquerade("name"), "success")

    def test_add_port(self):
        """
        Test adding a port to a specific zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(firewalld.add_port("zone", "80/tcp"), "success")

    def test_remove_port(self):
        """
        Test removing a port from a specific zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(firewalld.remove_port("zone", "80/tcp"), "success")

    def test_list_ports(self):
        """
        Test listing ports within a zone
        """
        ret = "22/tcp 53/udp 53/tcp"
        exp = ["22/tcp", "53/udp", "53/tcp"]

        with patch.object(firewalld, "__firewall_cmd", return_value=ret):
            self.assertEqual(firewalld.list_ports("zone"), exp)

    def test_add_port_fwd(self):
        """
        Test adding port forwarding on a zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(
                firewalld.add_port_fwd("zone", "22", "2222", "tcp"), "success"
            )

    def test_remove_port_fwd(self):
        """
        Test removing port forwarding on a zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(
                firewalld.remove_port_fwd("zone", "22", "2222", "tcp"), "success"
            )

    def test_list_port_fwd(self):
        """
        Test listing all port forwarding for a zone
        """
        ret = "port=23:proto=tcp:toport=8080:toaddr=\nport=80:proto=tcp:toport=443:toaddr="
        exp = [
            {
                "Destination address": "",
                "Destination port": "8080",
                "Protocol": "tcp",
                "Source port": "23",
            },
            {
                "Destination address": "",
                "Destination port": "443",
                "Protocol": "tcp",
                "Source port": "80",
            },
        ]

        with patch.object(firewalld, "__firewall_cmd", return_value=ret):
            self.assertEqual(firewalld.list_port_fwd("zone"), exp)

    def test_block_icmp(self):
        """
        Test ICMP block
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            with patch.object(firewalld, "get_icmp_types", return_value="echo-reply"):
                self.assertEqual(firewalld.block_icmp("zone", "echo-reply"), "success")

        with patch.object(firewalld, "__firewall_cmd"):
            self.assertFalse(firewalld.block_icmp("zone", "echo-reply"))

    def test_allow_icmp(self):
        """
        Test ICMP allow
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            with patch.object(firewalld, "get_icmp_types", return_value="echo-reply"):
                self.assertEqual(firewalld.allow_icmp("zone", "echo-reply"), "success")

        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertFalse(firewalld.allow_icmp("zone", "echo-reply"))

    def test_list_icmp_block(self):
        """
        Test ICMP block list
        """
        ret = "echo-reply echo-request"
        exp = ["echo-reply", "echo-request"]

        with patch.object(firewalld, "__firewall_cmd", return_value=ret):
            self.assertEqual(firewalld.list_icmp_block("zone"), exp)

    def test_get_rich_rules(self):
        """
        Test listing rich rules bound to a zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value=""):
            self.assertEqual(firewalld.get_rich_rules("zone"), [])

    def test_add_rich_rule(self):
        """
        Test adding a rich rule to a zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(
                firewalld.add_rich_rule(
                    "zone", 'rule family="ipv4" source address="1.2.3.4" accept'
                ),
                "success",
            )

    def test_remove_rich_rule(self):
        """
        Test removing a rich rule to a zone
        """
        with patch.object(firewalld, "__firewall_cmd", return_value="success"):
            self.assertEqual(
                firewalld.remove_rich_rule(
                    "zone", 'rule family="ipv4" source address="1.2.3.4" accept'
                ),
                "success",
            )



"""
Tests for the salt-run command
"""

import logging

import pytest

pytestmark = [
    pytest.mark.slow_test,
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def pillar_tree(base_env_pillar_tree_root_dir, salt_minion, salt_sub_minion, salt_cli):
    top_file = """
    base:
      '{}':
        - basic
      '{}':
        - basic
    """.format(
        salt_minion.id, salt_sub_minion.id
    )
    basic_pillar_file = """
    monty: python
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    basic_tempfile = pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    )
    try:
        with top_tempfile, basic_tempfile:
            ret = salt_cli.run("saltutil.refresh_pillar", wait=True, minion_tgt="*")
            assert ret.exitcode == 0
            assert salt_minion.id in ret.json
            assert ret.json[salt_minion.id] is True
            assert salt_sub_minion.id in ret.json
            assert ret.json[salt_sub_minion.id] is True
            yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = salt_cli.run("saltutil.refresh_pillar", wait=True, minion_tgt="*")
        assert ret.exitcode == 0
        assert salt_minion.id in ret.json
        assert ret.json[salt_minion.id] is True
        assert salt_sub_minion.id in ret.json
        assert ret.json[salt_sub_minion.id] is True


def test_cache(salt_run_cli):
    """
    Store, list, fetch, then flush data
    """
    # Store the data
    ret = salt_run_cli.run(
        "cache.store",
        bank="cachetest/runner",
        key="test_cache",
        data="The time has come the walrus said",
    )
    assert ret.exitcode == 0
    # Make sure we can see the new key
    ret = salt_run_cli.run("cache.list", bank="cachetest/runner")
    assert ret.exitcode == 0
    assert "test_cache" in ret.json

    # Make sure we can see the new data
    ret = salt_run_cli.run("cache.fetch", bank="cachetest/runner", key="test_cache")
    assert ret.exitcode == 0
    assert "The time has come the walrus said" in ret.stdout

    # Make sure we can delete the data
    ret = salt_run_cli.run("cache.flush", bank="cachetest/runner", key="test_cache")
    assert ret.exitcode == 0

    ret = salt_run_cli.run("cache.list", bank="cachetest/runner")
    assert ret.exitcode == 0
    assert "test_cache" not in ret.json


def test_cache_invalid(salt_run_cli):
    """
    Store, list, fetch, then flush data
    """
    ret = salt_run_cli.run("cache.store")
    assert ret.exitcode == 0
    # Make sure we can see the new key
    assert "Passed invalid arguments:" in ret.stdout


def test_grains(salt_run_cli, pillar_tree, salt_minion):
    """
    Test cache.grains
    """
    ret = salt_run_cli.run("cache.grains", tgt=salt_minion.id)
    assert ret.exitcode == 0
    assert salt_minion.id in ret.json


def test_pillar(salt_run_cli, pillar_tree, salt_minion, salt_sub_minion):
    """
    Test cache.pillar
    """
    ret = salt_run_cli.run("cache.pillar", tgt=salt_minion.id)
    assert ret.exitcode == 0
    assert salt_minion.id in ret.json
    assert salt_sub_minion.id not in ret.json


def test_pillar_no_tgt(salt_run_cli, pillar_tree, salt_minion, salt_sub_minion):
    """
    Test cache.pillar when no tgt is
    supplied. This should return pillar
    data for all minions
    """
    ret = salt_run_cli.run("cache.pillar")
    assert ret.exitcode == 0
    assert salt_minion.id in ret.json
    assert salt_sub_minion.id in ret.json


def test_pillar_minion_noexist(salt_run_cli, pillar_tree, salt_minion, salt_sub_minion):
    """
    Test cache.pillar when the target does not exist
    """
    ret = salt_run_cli.run("cache.pillar", tgt="non-exiting-minion")
    assert ret.exitcode == 0
    assert salt_minion.id not in ret.json
    assert salt_sub_minion.id not in ret.json


def test_pillar_minion_tgt_type_pillar(
    salt_run_cli, pillar_tree, salt_minion, salt_sub_minion
):
    """
    Test cache.pillar when the target exists
    and tgt_type is pillar
    """
    ret = salt_run_cli.run("cache.pillar", tgt="monty:python", tgt_type="pillar")
    assert ret.exitcode == 0
    assert salt_minion.id in ret.json
    assert salt_sub_minion.id in ret.json


def test_mine(salt_run_cli, salt_minion):
    """
    Test cache.mine
    """
    ret = salt_run_cli.run("cache.mine", tgt=salt_minion.id)
    assert ret.exitcode == 0
    assert salt_minion.id in ret.json



"""
Functions to translate input for network creation
"""

from salt.exceptions import SaltInvocationError

from . import helpers

ALIASES = {
    "driver_opt": "options",
    "driver_opts": "options",
    "ipv6": "enable_ipv6",
}
IPAM_ALIASES = {
    "ip_range": "iprange",
    "aux_address": "aux_addresses",
}
# ALIASES is a superset of IPAM_ALIASES
ALIASES.update(IPAM_ALIASES)
ALIASES_REVMAP = {y: x for x, y in ALIASES.items()}

DEFAULTS = {"check_duplicate": True}


def _post_processing(
    kwargs, skip_translate, invalid
):  # pylint: disable=unused-argument
    """
    Additional network-specific post-translation processing
    """
    # If any defaults were not expicitly passed, add them
    for item in DEFAULTS:
        if item not in kwargs:
            kwargs[item] = DEFAULTS[item]


# Functions below must match names of docker-py arguments
def driver(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def options(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter="=")


def ipam(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_dict(val)


def check_duplicate(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def internal(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def labels(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_labels(val)


def enable_ipv6(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def attachable(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def ingress(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


# IPAM args
def ipam_driver(val, **kwargs):  # pylint: disable=unused-argument
    return driver(val, **kwargs)


def ipam_opts(val, **kwargs):  # pylint: disable=unused-argument
    return options(val, **kwargs)


def ipam_pools(val, **kwargs):  # pylint: disable=unused-argument
    if not hasattr(val, "__iter__") or not all(isinstance(x, dict) for x in val):
        # Can't do a simple dictlist check because each dict may have more than
        # one element.
        raise SaltInvocationError("ipam_pools must be a list of dictionaries")
    skip_translate = kwargs.get("skip_translate", ())
    if not (skip_translate is True or "ipam_pools" in skip_translate):
        _globals = globals()
        for ipam_dict in val:
            for key in list(ipam_dict):
                if skip_translate is not True and key in skip_translate:
                    continue
                if key in IPAM_ALIASES:
                    # Make sure we resolve aliases, since this wouldn't have
                    # been done within the individual IPAM dicts
                    ipam_dict[IPAM_ALIASES[key]] = ipam_dict.pop(key)
                    key = IPAM_ALIASES[key]
                if key in _globals:
                    ipam_dict[key] = _globals[key](ipam_dict[key])
    return val


def subnet(val, **kwargs):  # pylint: disable=unused-argument
    validate_ip_addrs = kwargs.get("validate_ip_addrs", True)
    val = helpers.translate_str(val)
    if validate_ip_addrs:
        helpers.validate_subnet(val)
    return val


def iprange(val, **kwargs):  # pylint: disable=unused-argument
    validate_ip_addrs = kwargs.get("validate_ip_addrs", True)
    val = helpers.translate_str(val)
    if validate_ip_addrs:
        helpers.validate_subnet(val)
    return val


def gateway(val, **kwargs):  # pylint: disable=unused-argument
    validate_ip_addrs = kwargs.get("validate_ip_addrs", True)
    val = helpers.translate_str(val)
    if validate_ip_addrs:
        helpers.validate_ip(val)
    return val


def aux_addresses(val, **kwargs):  # pylint: disable=unused-argument
    validate_ip_addrs = kwargs.get("validate_ip_addrs", True)
    val = helpers.translate_key_val(val, delimiter="=")
    if validate_ip_addrs:
        for address in val.values():
            helpers.validate_ip(address)
    return val



"""
    :codeauthor: :email:`Simon Dodsley <simon@purestorage.com>`
"""


import salt.modules.purefb as purefb
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class PureFBTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.purefb
    """

    def setup_loader_modules(self):
        return {purefb: {}}

    def test_fs_create(self):
        """
        Test for creation of a filesystem
        """
        with patch.object(purefb, "fs_create", return_value=True):
            self.assertEqual(purefb.fs_create("test"), True)

    def test_fs_delete(self):
        """
        Test for deletion of a filesystem
        """
        with patch.object(purefb, "fs_delete", return_value=True):
            self.assertEqual(purefb.fs_delete("test"), True)

    def test_fs_eradicate(self):
        """
        Test for eradication of a filesystem
        """
        with patch.object(purefb, "fs_eradicate", return_value=True):
            self.assertEqual(purefb.fs_eradicate("test"), True)

    def test_fs_extend(self):
        """
        Test for size extention of a filesystem
        """
        with patch.object(purefb, "fs_extend", return_value=True):
            self.assertEqual(purefb.fs_extend("test", "33G"), True)

    def test_snap_create(self):
        """
        Test for creation of a filesystem snapshot
        """
        with patch.object(purefb, "snap_create", return_value=True):
            self.assertEqual(purefb.snap_create("test", suffix="suffix"), True)

    def test_snap_delete(self):
        """
        Test for deletion of a filesystem snapshot
        """
        with patch.object(purefb, "snap_delete", return_value=True):
            self.assertEqual(purefb.snap_delete("test", suffix="suffix"), True)

    def test_snap_eradicate(self):
        """
        Test for eradication of a deleted filesystem snapshot
        """
        with patch.object(purefb, "snap_eradicate", return_value=True):
            self.assertEqual(purefb.snap_eradicate("test", suffix="suffix"), True)
